﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Drawing;
using System.Diagnostics;

using RootPro.RootProCAD;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.UI;

namespace RCAddInSelectLayerGroupCS
{

    public partial class AddInTools
    {
        // RootPro CAD 作図編集ウィンドウのタブのタイプ
        public enum RC_TabViewType {
	        RC_TabViewType_Empty = 0,			// なし
	        RC_TabViewType_Paper = 1,			// 用紙
	        RC_TabViewType_Part = 2,			// 部分図
	        RC_TabViewType_DrawComponent = 3,	// 作図部品
            RC_TabViewType_LayerView = 4,       // レイヤ一覧
        };

        /// <summary>
        /// 作図編集ウィンドウのアクティブなタブビュータイプ取得
        /// </summary>
        /// <returns></returns>
        static public RC_TabViewType GetActiveTabViewType(AppAddIn app)
        {
            RC_TabViewType tabViewType = RC_TabViewType.RC_TabViewType_Empty;

            if(app.ActiveWindow == null)
                return tabViewType;

            TabView tabView = app.ActiveWindow.ActiveTabView;
            if(tabView == null)
                return tabViewType;

            TabViewType rcTabViewType = tabView.TabViewType;

            // 用紙ビューの場合
            if(rcTabViewType == TabViewType.PaperView)
                tabViewType = RC_TabViewType.RC_TabViewType_Paper;
            // 部分図ビューの場合
            else if(rcTabViewType == TabViewType.PartView)
                tabViewType = RC_TabViewType.RC_TabViewType_Part;
            // 作図部品編集ビューの場合
            else if(rcTabViewType == TabViewType.PartView)
                tabViewType = RC_TabViewType.RC_TabViewType_DrawComponent;
            // レイヤ一覧ビューの場合
            else if(rcTabViewType == TabViewType.LayerView)
                tabViewType = RC_TabViewType.RC_TabViewType_LayerView;

            return tabViewType;
        }

        /// <summary>
        /// 削除していいレイヤか
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="layer"></param>
        /// <param name="checkCurrentLayerInGroup">レイヤグループ内のカレントレイヤも対象にするか</param>
        /// <param name="useShapeLayerDeleteable">true:レイヤに図形があったっても削除可能とする</param>
        /// <param name="showMessage">メッセージを表示する</param>
        /// <returns></returns>
        static public bool IsDeleteableLayer(Document doc, Layer layer, bool checkCurrentLayerInGroup = true, bool showMessage = true)
        {
            // カレントレイヤは削除できない
            if(IsCurrentLayer(doc, layer, checkCurrentLayerInGroup))
            {
                if(showMessage)
                {
                    string msg = "";
                    if(!layer.IsGroup)
                        msg = String.Format(Properties.Resources.dontDeleteCurrentLayerMsg, layer.Name);// カレントレイヤなので削除できません。
                    else
                        msg = Properties.Resources.dontDeleteLayerGroupMsg;// "このレイヤグループにはカレントのレイヤが含まれているため削除できません。";

                    MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                }
                
                return false;
            }

            // 作図部品定義内のフィーチャで使用されているレイヤは削除不可
            if(IsComponentDefineIDUsingLayer(doc, layer))
            {
                if(showMessage)
                {
                    // "[" + layer.Name + "] " + "は作図部品定義で使われているため削除できません。";
                    string msg = msg = String.Format(Properties.Resources.dontDeleteDrawComponentLayerMsg, layer.Name);
                    MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                }
                 
                return false;
            }

            // 作図グループ内のフィーチャで使用されているレイヤは削除不可
            if(IsUsedDrawGroupLayer(doc, layer))
            {
                if(showMessage)
                {
                    // "[" + layer.Name + "] " + "は作図グループで使われているため削除できません。";
                    string msg = String.Format(Properties.Resources.dontDeleteDrawGroupLayerMsg, layer.Name);
                    MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                }
                
                return false;
            }

            return true;
        }

        /// <summary>
        /// 指定したレイヤやレイヤグループが図形で使われているか。使われている場合削除していいか確認メッセージを表示する
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="layer"></param>
        /// <returns></returns>
        static public bool CheckAndQuestionDeleteUseLayerShape(Document doc, Layer layer)
        {
            // レイヤ上に図形があれば確認
            if(IsUseLayerShape(doc, layer))
            {
                string msg = String.Format(Properties.Resources.dontDeleteDrawGroupLayerMsg, layer.Name);//"[" + layer.Name + "] " + "は図形が含まれています。";
                if(!layer.IsGroup)
                    msg += "\n" + Properties.Resources.deleteUseLayerQuestion;// "このレイヤを削除しますか?";
                else
                    msg += "\n" + Properties.Resources.deleteUseLayerGroupQuestion;// "このレイヤグループを削除しますか?";

                DialogResult dialogResult = MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2);
                if(dialogResult != DialogResult.Yes)
                    return false;
            }

            return true;
        }

        /// <summary>
        /// ルート直下にレイヤがあるか
        /// レイヤグループ以外
        /// </summary>
        /// <param name="doc"></param>
        /// <returns></returns>
        static public bool IsRootLayers(Document doc)
        {
            LayerTable layerTable = doc.LayerTable;
            LayerCollection layers = layerTable.RootLayer.ChildLayers;
            foreach(Layer layer in layers)
            {
                // レイヤグループ以外
                if(!layer.IsGroup)
                    return true;
            }

            return false;
        }

        /// <summary>
        /// ルートに指定したID以外のレイヤがあるか
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="layerID">-1:ルートに1つでもあればtrueで返す</param>
        /// <returns></returns>
        static public bool IsOtherRootLayers(Document doc, int layerID)
        {
            LayerTable layerTable = doc.LayerTable;
            LayerCollection layers = layerTable.RootLayer.ChildLayers;
            foreach(Layer layer in layers)
            {
                // レイヤグループ以外で指定したレイヤID以外
                if(!layer.IsGroup && layer.ID != layerID)
                    return true;
            }

            return false;
        }

        /// <summary>
        /// 指定したレイヤまたはレイヤグループ内のレイヤがカレントか
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="layer"></param>
        /// <param name="checkCurrentLayerInGroup">レイヤグループ内のカレントレイヤも対象にするか</param>
        /// <returns></returns>
        static public bool IsCurrentLayer(Document doc, Layer layer, bool checkCurrentLayerInGroup = true)
        {
            // レイヤグループの場合
            if(checkCurrentLayerInGroup && layer.IsGroup)
            {
                // 子レイヤについて調べる
                LayerCollection layers = layer.ChildLayers;
                foreach(Layer childLayer in layers)
                {
                    // 再帰処理
                    if(IsCurrentLayer(doc, childLayer, checkCurrentLayerInGroup))
                        return true;
                }
            }
            // レイヤの場合
            else
            {
                if(layer.IsCurrent)
                    return true;
            }

            return false;
        }

        /// <summary>
        /// 指定したレイヤ名で最初に見つかったレイヤ取得
        /// </summary>
        /// <param name="layers"></param>
        /// <param name="layerName"></param>
        /// <returns></returns>
        static public Layer GetFirstFindedLayerByName(LayerCollection layers, string layerName)
        {

            foreach(Layer childLayer in layers)
            {
                if(childLayer.IsGroup)
                {
                    // 再帰処理
                    Layer layer = GetFirstFindedLayerByName(childLayer.ChildLayers, layerName);
                    if(layer != null)
                        return layer;
                }
                else
                {
                    if(childLayer.Name == layerName)
                        return childLayer;
                }
            }

            return null;
        }

        /// <summary>
        /// 指定したレイヤまたはレイヤグループ内のレイヤが作図部品定義内のフィーチャで使用されているか
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="layoutId"></param>
        /// <returns></returns>
        static public bool IsComponentDefineIDUsingLayer(Document doc, Layer layer)
        {
            // レイヤグループの場合
            if(layer.IsGroup)
            {
                // 子レイヤについて調べる
                LayerCollection layers = layer.ChildLayers;
                foreach(Layer childLayer in layers)
                {
                    // 再帰処理
                    if(IsComponentDefineIDUsingLayer(doc, childLayer))
                        return true;
                }
            }
            // レイヤの場合
            else
            {
                // 作図部品定義でレイヤIDが使用されているか？
                MasterDrawComponentCollection components = doc.MasterDrawComponents;
                foreach(MasterDrawComponent component in components)
                {
                    MasterDrawComponentShapeItemCollection shapes = component.Shapes;
                    foreach(Shape shape in shapes)
                    {
                        if(shape.LayerID == layer.ID)
                            return true;
                    }
                }
            }
            
	        return false;
        }

        /// <summary>
        /// 指定したレイヤまたはレイヤグループ内のレイヤがドキュメント中の作図グループ内のフィーチャで使用されているか
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="layer"></param>
        /// <returns></returns>
        static public bool IsUsedDrawGroupLayer(Document doc, Layer layer)
        {
            // レイヤグループの場合
	        if(layer.IsGroup)
	        {
                // 子レイヤについて調べる
                LayerCollection layers = layer.ChildLayers;
                foreach(Layer childLayer in layers)
		        {
                    // 再帰処理
                    if(IsUsedDrawGroupLayer(doc, childLayer))
                        return true;
		        }
	        }
            // レイヤの場合
            else
            {
                // 全部分図から
                DrawingCollection drawings = doc.Drawings;
                foreach(Drawing drawing in drawings)
                {
                    // 全図形から
                    ShapeCollection shapes = drawing.Shapes;
                    foreach(Shape shape in shapes)
                    {
                        // 作図グループだった場合
                        if(shape is DrawGroupShape)
                        {
                            if(IsUsedDrawGroupLayer((DrawGroupShape)shape, layer))
                                return true;
                        }
                    }
                }
            }

            return false;
        }
        
        /// <summary>
        /// 指定したレイヤが指定した作図グループで使われているか
        /// </summary>
        /// <param name="drawGroupShape"></param>
        /// <param name="layer"></param>
        /// <returns></returns>
        static private bool IsUsedDrawGroupLayer(DrawGroupShape drawGroupShape, Layer layer)
        {
            DrawGroupShapeItemCollection chaildShapes = drawGroupShape.Shapes;

            foreach(Shape shape in chaildShapes)
            {
                 // 作図グループだった場合
                 if(shape is DrawGroupShape)
                 {
                    // 再帰処理
                    if(IsUsedDrawGroupLayer((DrawGroupShape)shape, layer))
                        return true;
                 }
                 else
                 {
                    if(shape.LayerID == layer.ID)
                        return true;
                 }
            }

            return false;
        }

        /// <summary>
        /// 指定したレイヤまたはレイヤグループ内のレイヤがドキュメント中の図形で使われているか
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="layer"></param>
        /// <param name="withoutDrawGroup">true:作図グループは除く</param>
        /// <returns></returns>
        static public bool IsUseLayerShape(Document doc, Layer layer, bool withoutDrawGroup = false)
        {
            // レイヤグループの場合
            if(layer.IsGroup)
            {
                // 子レイヤについて調べる
                LayerCollection layers = layer.ChildLayers;
                foreach(Layer childLayer in layers)
                {
                    // 再帰処理
                    if(IsUseLayerShape(doc, childLayer, withoutDrawGroup))
                        return true;
                }
            }
            // レイヤの場合
            else
            {
                // 全部分図から
                DrawingCollection drawings = doc.Drawings;
                foreach(Drawing drawing in drawings)
                {
                    // 全図形から
                    ShapeCollection shapes = drawing.Shapes;
                    foreach(Shape shape in shapes)
                    {
                        // 作図グループだった場合
                        if(shape is DrawGroupShape)
                        {
                            if(!withoutDrawGroup)
                                if(IsUsedDrawGroupLayer((DrawGroupShape)shape, layer))
                                    return true;
                        }
                        else
                        {
                            if(shape.LayerID == layer.ID)
                                return true;
                        }
                    }
                }
            }

            return false;
        }

        /// <summary>
        /// 用紙上で部分図レイアウトがカレントか
        /// </summary>
        /// <returns></returns>
        static public  bool IsCurrentDrawingLayoutOnPaper(AppAddIn app)
        {
            // 作図編集ウィンドウの用紙タブがアクティブでなければfalse
            RC_TabViewType tabViewType = GetActiveTabViewType(app);
            if(tabViewType != RC_TabViewType.RC_TabViewType_Paper)
                return false;

            Document doc = app.ActiveDocument;
            Debug.Assert(doc != null);
            if(doc == null)
                return false;

            Layout currentLayout = doc.CurrentLayout;
            if(currentLayout != null)
            {
                // レイアウト名
                string layoutName = currentLayout.Name;
                if(layoutName != "")
                    return true;
            }

            return false;
        }

        /// <summary>
        /// 用紙の部分図レイアウト取得(複数用紙がある場合は最初の用紙の)
        /// </summary>
        /// <param name="doc"></param>
        /// <returns></returns>
        static public Layout GetPaperDrawingLayout(Document doc)
        {
            Debug.Assert(doc != null);
 
            DrawingCollection drawings = doc.Drawings;
            foreach(Drawing drawing in drawings)
            {
                LayoutCollection layouts = drawing.Layouts;
                foreach(Layout layout in layouts)
                {
                    if(layout.Drawing is Paper)
                        return layout;
                }
            }

            return null;
        }



        /// <summary>
        /// 指定したIDの部分図レイアウトの取得
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        static public Layout GetLayoutOfID(Document doc, int layoutId)
        {
            Layout layout = null;

            if(doc == null)
                return layout;

            if(layoutId >= 0)
            {
                DrawingCollection drawings = doc.Drawings;
                foreach(Drawing drawing in drawings)
                {
                    LayoutCollection layouts = drawing.Layouts;
                    if(layouts.Count == 0)
                        continue;

                    // 指定したIDの部分図レイアウトがあればそれを返す
                    int index = layouts.IndexOfID(layoutId);
                    if(index >= 0)
                    {
                        layout = layouts[index];
                        return layout;
                    }
                }
            }

            return layout;
        }

        /// <summary>
        /// 部分図レイアウトの内で用紙のもののインデックスを取得
        /// </summary>
        /// <param name="layouts"></param>
        /// <returns></returns>
        static public int GetPaperDrawingLayoutIndex(LayoutCollection layouts)
        {
            int layoutsCount = layouts.Count;
            for(int i = 0; i < layoutsCount; i++)
            {
                Layout layout = layouts[i];
                if(layout.Drawing is Paper)
                    return i;
            }

            return -1;
        }

        /// <summary>
        /// 有効なレイヤIDか
        /// ルート直下のレイヤーグループ内のレイヤに一致したIDのレイヤがあるか
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="layerId"></param>
        /// <returns></returns>
        static public bool IsValidLayerID(Document doc, int layerId)
        {
            LayerTable layerTable = doc.LayerTable;

            LayerCollection layerCollection = layerTable.RootLayer.ChildLayers;
            foreach(Layer layer in layerCollection)
            {
                // レイヤグループの場合
                if(layer.IsGroup)
                {
                    // 子レイヤについて調べる
                    LayerCollection layers = layer.ChildLayers;
                    foreach(Layer childLayer in layers)
                    {
                        if(!childLayer.IsGroup && childLayer.ID == layerId)
                            return true;
                    }
                }
            }

            return false;
        }

        /// <summary>
        /// 新しいレイヤグループの名前取得
        /// ルート直下のレイヤグループに新しく追加する名前(「グループ 1」～ 「グループ A」, 「グループ B」,「グループ C」...)
        /// すべての名前が既に使われている場合は""
        /// </summary>
        /// <param name="layerTable"></param>
        /// <param name="maxLayerGroupSize">レイヤグループの最大数</param>
        /// <returns></returns>
        static public string GetNewLayerGroupName(LayerTable layerTable, int maxLayerGroupSize)
        {
            string newLayerGroupName = "";

            for(int i = 0; i < maxLayerGroupSize; i++)
            {
                string newName = "Group " + i.ToString("X"); // 16進数文字で
                if(!IsLayerGroupName(layerTable, newName))
                {
                    newLayerGroupName = newName;
                    break;
                }
            }

            return newLayerGroupName;
        }

        /// <summary>
        /// 同じ名前のレイヤグループがあるか
        /// ルート直下のレイヤグループが対象
        /// </summary>
        /// <param name="layerTable"></param>
        /// <param name="name"></param>
        /// <returns></returns>
        static public bool IsLayerGroupName(LayerTable layerTable, string name)
        {
            LayerCollection layerCollection = layerTable.RootLayer.ChildLayers;
            foreach(Layer layer in layerCollection)
            {
                // レイヤグループの場合
                if(layer.IsGroup)
                {
                    if(layer.Name == name)
                        return true;
                }
            }

            return false;
        }

        /// <summary>
        /// ルートにあるレイヤグループの数を取得
        /// </summary>
        /// <param name="layerTable"></param>
        /// <returns></returns>
        static public int GetRootLayerGroupCount(LayerTable layerTable)
        {
            int rootLayerGroupCount = 0;

            LayerCollection layerCollection = layerTable.RootLayer.ChildLayers;
            foreach(Layer layer in layerCollection)
            {
                // レイヤグループの場合
                if(layer.IsGroup)
                    rootLayerGroupCount++;
            }

            return rootLayerGroupCount;
        }

        /// <summary>
        /// 指定したレイヤグループ以外の前後のレイヤグループのレイヤを探す
        /// 前のレイヤグループのレイヤ優先
        /// </summary>
        /// <param name="layerTable"></param>
        /// <param name="thisLayerGroup"></param>
        /// <returns></returns>
        static public Layer FindOtherLayerGroupsLayer(LayerTable layerTable, Layer thisLayerGroup)
        {
            Layer otherLayer = null;

            bool isThisLayerGroup = false;
            LayerCollection rootLayers = layerTable.RootLayer.ChildLayers;
            foreach(Layer layer in rootLayers)
            {
                // レイヤグループの場合
                if(layer.IsGroup)
                {
                    // 指定したレイヤグループの場合
                    if(layer.Equals(thisLayerGroup))
                    {
                        // 前後を分けるためのフラグOn
                        isThisLayerGroup = true;
                        continue;
                    }
                    // 指定したレイヤグループ以外
                    else
                    {
                        LayerCollection childLayers = layer.ChildLayers;
                        foreach(Layer childLayer in childLayers)
                        {
                            // レイヤの場合
                            if(!childLayer.IsGroup)
                            {
                                // 指定したレイヤグループの後の場合
                                if(isThisLayerGroup)
                                {
                                    // 先のグループにレイヤがあった場合はここで返す
                                    if(otherLayer != null)
                                        return otherLayer;
                                    // 先のレイヤグループにレイヤがなければこのレイヤを返す
                                    if(otherLayer == null)
                                        return childLayer;
                                }
                                // 指定したレイヤグループより前のレイヤーグループの最初のレイヤを保持して次を探す
                                else
                                {
                                    otherLayer = childLayer;
                                    break;
                                }
                            }
                        }
                    }
                }
            }

            return otherLayer;
        }

        /// <summary>
        /// 新しい部分図名を取得
        /// </summary>
        /// <param name="drawings"></param>
        /// <param name="maxCount"></param>
        /// <returns></returns>
        static public string GetNewDrawingName(DrawingCollection drawings, int maxCount)
        {
            string newLayoutName = "";

            for(int i = 0; i < maxCount; i++)
            {
                string newName = "Part " + i.ToString("X"); // 16進数文字で

                bool exist = false;
                foreach(Drawing drawing in drawings)
                {
                    if(drawing.Name == newName)
                    {
                        exist = true;
                        break;
                    }
                }

                if(!exist)
                {
                    newLayoutName = newName;
                    break;
                }
            }

            return newLayoutName;
        }

        /// <summary>
        /// コントロールを名前で探す
        /// </summary>
        /// <param name="parent"></param>
        /// <param name="controlName"></param>
        /// <returns></returns>
        static public Control FindControl(Control parent, string controlName)
        {
            // parent のコントロールを列挙する
            foreach (Control control in parent.Controls)
            {
                // 子のコントロールがあった場合、再帰呼び出し
                if(control.HasChildren)
                {
                    Control findControl = FindControl(control, controlName);
                    if(findControl != null)
                    {
                        return findControl;
                    }
                }

                // コントロール名が同じ場合はそのコントロールを返す
                if(control.Name == controlName)
                   return control;
            }

            return null;
        }

    }
}