﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Drawing;
using System.Diagnostics;

using RootPro.RootProCAD;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.Command;


namespace RCAddInSelectLayerGroupCS
{
    public partial class RootProCADUserControl : UserControl
    {
        /// <summary>
        /// ドキュメントのチェック
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="strErr"></param>
        /// <returns></returns>
        bool IsConvertableDocument(Document doc, out string strErr)
        {
            int err = 0;
            strErr = "";

            Cursor preCursor = Cursor.Current;
            Cursor.Current = Cursors.WaitCursor;// 待機カーソルに

            // 以下の条件を満たせばよいとする。
            // ・用紙は1つ
            // ・部分図レイアウト内にのみ図形があり、用紙上に図形はない 
            // ・部分図レイアウトは 16 以内 （用紙自身は除く）--- 推奨するが特にチェックしていない
            // ・同じ部分図を元にした複数の部分図レイアウトはない 
            // ・部分図レイアウトは XY 同一スケール 
            // ・部分図レイアウトの配置角度は 0 度 
            // ・レイヤはルートになく、1 階層までのレイヤグループが 16 個以内で、各レイヤグループには 16 個以内のレイヤがある 
            // ・同じレイヤを異なる部分図の図形で使っていない

            int paperCount = 0;
            DrawingCollection drawings = doc.Drawings;
            foreach(Drawing drawing in drawings)
            {
                if(drawing is Paper)
                {
                    paperCount++;

                    // 用紙が複数
                    if(paperCount > 1)
                    {
                        err = 1;
                        goto ERRLABEL;
                    }

                    // 部分図レイアウト
                    LayoutCollection layouts = drawing.Layouts;
                    if(layouts.Count == 0)
                        continue;

                    List<int> drawingIDList = new List<int>();
                    foreach(Layout layout in layouts)
                    {
                        // 用紙自身のはスキップ
                        if(layout.Drawing is Paper)
                            continue;

                        // 元の部分図のID
                        int layoutDrawingID = layout.Drawing.ID;
                        // 同じ部分図を元にした複数の部分図レイアウトは1つ
                        if(drawingIDList.Contains(layoutDrawingID))
                        {
                            err = 3;
                            goto ERRLABEL;
                        }
                        drawingIDList.Add(layoutDrawingID);

                        // XY同一スケールではない場合
                        if(layout.EnableScaleY)
                        {
                            err = 4;
                            goto ERRLABEL;
                        }

                        // 配置角度が0以外
                        if(Math.Abs(0.0 - layout.Angle) > RcPrecisionAngular * (180d / Math.PI))
                        {
                            err = 5;
                            goto ERRLABEL;
                        }
                    }
                }

            }

            // 用紙がない
            if(paperCount == 0)
            {
                err = 2;
                goto ERRLABEL;
            }

            // レイヤはルートになく、1 階層までのレイヤグループが 16 個以内で、各レイヤグループには 16 個以内のレイヤがある 
            // レイヤテーブルの取得
            LayerTable layerTable = doc.LayerTable;

            // 子レイヤのレイヤコレクション取得
            LayerCollection layerCollection = layerTable.RootLayer.ChildLayers;
            int layerGroupCount = 0;
            foreach(Layer layer in layerCollection)
            {
                if(layer.IsGroup)
                {
                    // レイヤグループが多すぎる
                    layerGroupCount++;
                    if(layerGroupCount > RcAddInMaxLayerGroupSize)
                    {
                        err = 9;
                        goto ERRLABEL;
                    }

                    int childLayerCount = 0;
                    LayerCollection childLayers = layer.ChildLayers;
                    foreach(Layer childLayer in childLayers)
                    {
                        // レイヤグループに更にレイヤグループ
                        if(childLayer.IsGroup)
                        {
                            err = 7;
                            goto ERRLABEL;
                        }

                        // レイヤの数が多すぎる
                        childLayerCount++;
                        if(childLayerCount > RcAddInMaxChildLayerSize)
                        {
                            err = 8;
                            goto ERRLABEL;
                        }
                    }
                }
                // ルートにレイヤ
                else
                {
                    err = 6;
                    goto ERRLABEL;
                }
            }

            // 同じレイヤを異なる部分図の図形で使っていないか
            if(IsLayerUsedDifferentDrawing(doc, layerTable.RootLayer))
            {
                  err = 10;
                  goto ERRLABEL;
            }


            ERRLABEL:
            if(err != 0)
            {
                if(err == 1)// この図面には用紙は複数あります。
                    strErr = Properties.Resources.checkDocumentErrMsg_1;
                else if(err == 2)// この図面には用紙がありません。
                    strErr = Properties.Resources.checkDocumentErrMsg_2;
                else if(err == 3)// 同じ部分図を元にした複数の部分図レイアウトがあります。
                    strErr = Properties.Resources.checkDocumentErrMsg_3;
                else if(err == 4)// 部分図レイアウトに XY 同一スケールではないものがあります。
                    strErr = Properties.Resources.checkDocumentErrMsg_4;
                else if(err == 5)// 部分図レイアウトに配置角度が 0 度以外のものがあります。
                    strErr = Properties.Resources.checkDocumentErrMsg_5;
                else if(err == 6)// ルートにレイヤがあります。
                    strErr = Properties.Resources.checkDocumentErrMsg_6;
                else if(err == 7)// 2 階層以上のレイヤグループがあります。
                    strErr = Properties.Resources.checkDocumentErrMsg_7;
                else if(err == 8)// レイヤグループ内のレイヤの数が {0} を超えています。
                    strErr = String.Format(Properties.Resources.checkDocumentErrMsg_8, RcAddInMaxChildLayerSize);
                else if(err == 9)// レイヤグループの数が {0} を超えています。
                    strErr = String.Format(Properties.Resources.checkDocumentErrMsg_9, RcAddInMaxLayerGroupSize);
                else if(err == 10)// 同じレイヤが異なる部分図の図形で使われています。
                    strErr = Properties.Resources.checkDocumentErrMsg_10;
                else
                    strErr = "";
            }

            Cursor.Current = preCursor;// カーソルを元に戻す

            bool stat = err == 0 ? true:false;
            return stat;
        }

        /// <summary>
        /// 同じレイヤが異なる部分図の図形で使われていないか
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="layerGroup"></param>
        /// <returns></returns>
        private bool IsLayerUsedDifferentDrawing(Document doc, Layer layerGroup)
        {
            Debug.Assert(doc != null);
            Debug.Assert(layerGroup != null);
            Debug.Assert(layerGroup.IsGroup);

            // レイヤ毎
            LayerCollection layerCollection = layerGroup.ChildLayers;
            foreach(Layer layer in layerCollection)
            {
                // レイヤグループだった場合
                if(layer.IsGroup)
                {
                    // 再帰処理
                    if(IsLayerUsedDifferentDrawing(doc, layer))
                        return true;
                }
                else
                {
                    // 異なる部分図で指定したレイヤが使われていないか全図形から調べる
                    int layerID = layer.ID;
                    int usedDrawingID = -1;// 最初に見つかった部分図のID(初期値-1)
                    // 部分図ごと
                    DrawingCollection drawings = doc.Drawings;
                    foreach(Drawing drawing in drawings)
                    {
                        // 部分図内で使われている全図形
                        ShapeCollection shapeCollection = drawing.Shapes;
                        foreach(Shape shape in shapeCollection)
                        {
                            // この図形でレイヤが使われている
                            if(shape.LayerID == layerID)
                            {
                                // 最初に見つかった部分図
                                if(usedDrawingID == -1)
                                {
                                    usedDrawingID = drawing.ID;
                                    break;
                                }
                                // 他の部分図でも使われていた
                                else
                                {
                                    return true;
                                }
                            }
                        }
                    }
                }
            }

            return false;
        }

    }
}
