﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Diagnostics;

using RootPro.RootProCAD;

namespace RCAddInSelectLayerGroupCS
{
    public partial class EditGroupNameForm : Form
    {
        Document doc = null;
        Layer layer = null;
        Layout layout = null;

        public EditGroupNameForm(Document doc, Layer layer, Layout layout)
        {
            this.doc = doc;
            this.layer = layer;
            this.layout = layout;

            InitializeComponent();
        }

        private void EditGroupNameForm_Load(object sender, EventArgs e)
        {
            layerGroupNameTextBox.Text = layer.Name;

            if(layout != null)
            {
                drawingNameTextBox.Text = layout.Drawing.Name;
            }
 
            UpdateControls();
        }

        private void drawingNameEqualLayerGroupNameCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            UpdateControls();
        }

        private void UpdateControls()
        {
            if(layout != null)
            {
                // 部分図の名前はレイヤグループと同じにする場合
                if(drawingNameEqualLayerGroupNameCheckBox.Checked)
                {
                    drawingNameLabel.Visible = false;
                    drawingNameTextBox.Visible = false;

                    commentLabel.Visible = true;
                }
                else
                {
                    drawingNameLabel.Visible = true;
                    drawingNameTextBox.Visible = true;

                    commentLabel.Visible = true;
                }
            }
            else
            {
                drawingNameEqualLayerGroupNameCheckBox.Visible = false;
                drawingNameLabel.Visible = false;
                drawingNameTextBox.Visible = false;
                commentLabel.Visible = false;
            }
        }

        /// <summary>
        /// レイヤーグループの名前取得
        /// </summary>
        /// <returns></returns>
        public string GetLayerGroupName()
        {
            string name = layerGroupNameTextBox.Text;
            //name.Trim();
            return name;
        }

        /// <summary>
        /// 部分図名の取得
        /// </summary>
        /// <returns></returns>
        public string GetDrawingName()
        {
            string name = drawingNameTextBox.Text;
            //name.Trim();
            return name;
        }

        /// <summary>
        /// 部分図レイアウトの名前
        /// 部分図名が変わる場合は部分図レイアウト名は"Layout(部分図名)"になる
        /// </summary>
        /// <returns></returns>
        public string GetDrawingLayoutName()
        {
            string name = "";

            if(this.layout != null)
            {
                Debug.Assert(drawingNameTextBox.Text != "");

                string oldDrawingName = layout.Drawing.Name;
                string drawingName = GetDrawingName();

                // 部分図名が変わる場合は部分図レイアウト名は"Layout(部分図名)"にする
                if(oldDrawingName != drawingName)
                    name = "Layout(" + drawingName + ")";
                else
                    name = layout.Name;
            }

            return name;
        }

        /// <summary>
        /// 同じ名前のレイヤ又はレイヤグループがないか
        /// </summary>
        /// <param name="parentID"></param>
        /// <param name="layerName"></param>
        /// <param name="myLayer"></param>
        /// <returns></returns>
        private bool IsExistLayerName(int parentID, string layerName, Layer myLayer)
        {
            Debug.Assert(this.layer != null);

            LayerTable layerTable = doc.LayerTable;
            Layer layerGroup = layerTable.GetLayerByID(parentID);
            Debug.Assert(layerGroup.IsGroup);
            LayerCollection layerCollection = layerGroup.ChildLayers;

            foreach(Layer chaildLayer in layerCollection)
            {
		        // 自身は無視する
		        if(chaildLayer.ID == myLayer.ID)
			        continue;

                // myLayerがグループではない場合グループは無視する
                if(!myLayer.IsGroup && chaildLayer.IsGroup)
                    continue;
                // myLayerがグループの場合グループでないレイヤは無視する
                else if(myLayer.IsGroup && !chaildLayer.IsGroup)
                    continue;

                // 大文字と小文字を区別しないで文字列を比較
                if(String.Compare(layerName, chaildLayer.Name, true) == 0)
			        return true;
	        }

	        return false;
        }

        /// <summary>
        /// 同じ名前の部分図がないか
        /// </summary>
        /// <param name="drawingName"></param>
        /// <param name="myDrawing"></param>
        /// <returns></returns>
        private bool IsExistDrawingName(string drawingName, Drawing myDrawing)
        {
            DrawingCollection drawings = doc.Drawings;
            foreach(Drawing drawing2 in drawings)
            {
                // 自身は無視する
                if(drawing2.ID == myDrawing.ID)
                    continue;

                // 大文字と小文字を区別しないで文字列を比較
                if(String.Compare(drawingName, drawing2.Name, true) == 0)
                    return true;
            }

            return false;
        }

        /// <summary>
        /// 同じ部分図レイアウトがないか
        /// </summary>
        /// <param name="drawingLayoutName"></param>
        /// <param name="mylayout"></param>
        /// <returns></returns>
        private bool IsExistDrawingLayoutName(string drawingLayoutName, Layout mylayout)
        {
            Debug.Assert(mylayout != null);
            Debug.Assert(mylayout.ParentDrawing != null);

            LayoutCollection layouts = mylayout.ParentDrawing.Layouts;
            foreach(Layout layer in layouts)
            {
                // 自身は無視する
                if(layer.ID == mylayout.ID)
                    continue;

                // 大文字と小文字を区別しないで文字列を比較
                if(String.Compare(drawingLayoutName, layer.Name, true) == 0)
                    return true;
            }

            return false;
        }

        /// <summary>
        /// OK ボタン
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void okButton_Click(object sender, EventArgs e)
        {
            // 部分図名はレイヤグループ名と同じにする場合
            if(drawingNameEqualLayerGroupNameCheckBox.Visible && drawingNameEqualLayerGroupNameCheckBox.Checked)
                drawingNameTextBox.Text = layerGroupNameTextBox.Text;

            if(this.layer != null)
            {
                // レイヤグループ名
                string layerGroupName = GetLayerGroupName();
                bool err = false;
                if(layerGroupName == "")
                {
                    err = true;
                }
                else
                {
                    int parentID = this.layer.ParentLayer.ID;
                    // 同じ名前のレイヤグループがないか名前チェック
                    bool isExistLayerName = IsExistLayerName(parentID, layerGroupName, this.layer);
                    if(isExistLayerName == true)
                        err = true;
                }

                if(err)
                {
                    string strErr = Properties.Resources.renameLayerGroupMsg;// 指定したレイヤグループ名と同じグループで既にあるか名前が不正です。
                    MessageBox.Show(strErr, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);

                    layerGroupNameTextBox.Focus();
                    layerGroupNameTextBox.SelectAll();
                    return;
                }
            }

            if(this.layout != null)
            {
                // 部分図名
                string drawingName = GetDrawingName();
                bool err = false;
                if(drawingName == "")
                {
                    err = true;
                }
                else
                {
                    // 同じ名前の部分図がないか名前チェック
                    bool isExistName = IsExistDrawingName(drawingName, this.layout.Drawing);
                    if(!isExistName)
                    {
                        // 同じ名前の部分図レイアウトがないか名前チェック
                        string drawingLayoutName = GetDrawingLayoutName();
                        isExistName = IsExistDrawingLayoutName(drawingLayoutName, this.layout);
                    }

                    if(isExistName == true)
                        err = true;
                }

                if(err)
                {
                    string strErr = Properties.Resources.renamePartDrawingMsg;// 指定した部分図名が既に使われているか名前が不正です。
                    MessageBox.Show(strErr, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);

                    drawingNameTextBox.Focus();
                    drawingNameTextBox.SelectAll();
                    return;
                }
            }

            this.DialogResult = DialogResult.OK;
            this.Close();
        }

        /// <summary>
        /// キャンセル ボタン
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void cancelButton_Click(object sender, EventArgs e)
        {
            this.DialogResult = DialogResult.Cancel;
            this.Close();
        }


    }
}
