﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Drawing;
using System.Diagnostics;

using RootPro.RootProCAD;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.Command;


namespace RCAddInSelectLayerGroupCS
{

    partial class RootProCADUserControl:UserControl
    {

        /// <summary>
        /// レイヤグループに部分図レイアウトを関連付けする
        /// レイヤグループ内のレイヤの先頭をグループ内のカレントレイヤともしておく
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="undoUnitFlg">UndoUnit で記録するか</param>
        /// <returns></returns>
        private bool InterlockingDrawinglayoutToLayerGroup(Document doc, bool undoUnitFlg = true)
        {
            Debug.Assert(doc != null);
            bool stat = false;

            // Jww ファイルの読み込みでは、作成された部分図の名前と同じ名前のレイヤーグループができる。
            // よって、レイヤグループ毎に同じ名前の部分図を探して、その部分図を元にした用紙上の部分図レイアウトを関連付けする

            Cursor preCursor = Cursor.Current;
            Cursor.Current = Cursors.WaitCursor;// 待機カーソルに

            // 用紙を取得
            Drawing paperDrawing = null;
            int paperCount = 0;
            DrawingCollection drawings = doc.Drawings;
            foreach(Drawing drawing in drawings)
            {
                if(drawing is Paper)
                {
                    paperCount++;
                    Debug.Assert(paperCount == 1);
                    if(paperCount > 1)
                        goto ENDLABEL;

                    paperDrawing = drawing;
                }
            }

            Debug.Assert(paperDrawing != null);
          
            // レイヤグループと同じ名前の部分図を使った部分図レイアウトをレイヤグループに関連付ける
            bool beginUndoUnitAtThisFunction = false;
            LayerTable layerTable = doc.LayerTable;
            int documentCurrentLayerID = -1;
            if(layerTable.CurrentLayer != null)
                documentCurrentLayerID = layerTable.CurrentLayer.ID;

            LayerCollection layerCollection = layerTable.RootLayer.ChildLayers;
            foreach(Layer layer in layerCollection)
            {
                if(layer.IsGroup)
                {
                    Layer layerGroup = layer;
                    string layerGroupName = layerGroup.Name;

                    LayoutCollection paperLayouts = paperDrawing.Layouts;
                    foreach(Layout layout in paperLayouts)
                    {
                        // 元の部分図の名前がレイヤグループと同じ部分図レイアウトなら
                        if(layout.Drawing.Name == layerGroupName)
                        {
                            // UndoUnitの開始
                            if(undoUnitFlg && !beginUndoUnitAtThisFunction)
                            {
                                doc.UndoManager.BeginUndoUnit();
                                beginUndoUnitAtThisFunction = true;
                            }
                           
                            // レイヤグループの説明にレイヤグループのカレントレイヤIDを書き込む
                            int currentLayerIDinLayerGroup = GetCurrentLayerIDinLayerGroup(layerGroup, documentCurrentLayerID);
                            if (currentLayerIDinLayerGroup >= 0)
                                SetIdToLayerGroupDescription(layerGroup, RootProCADUserControl.ItemName_CurrentLayerID, currentLayerIDinLayerGroup);
                            
                            // レイヤグループの説明に部分図レイアウトIDを書き込む
                            SetIdToLayerGroupDescription(layerGroup, RootProCADUserControl.ItemName_InterlockingDrawingLayoutID, layout.ID);
                        }
                    }
                }
            }

            // UndoUnitの終了
            if(beginUndoUnitAtThisFunction)
                doc.UndoManager.EndUndoUnit();

            stat = true;

            ENDLABEL:
            Cursor.Current = preCursor;// カーソルを元に戻す

            return stat;
        }

        /// <summary>
        /// レイヤグループ内でカレントにするレイヤのID取得
        /// </summary>
        /// <param name="layerGroup"></param>
        /// <param name="documentCurrentLayerID"></param>
        /// <returns></returns>
        private int GetCurrentLayerIDinLayerGroup(Layer layerGroup, int documentCurrentLayerID)
        {
            Debug.Assert(layerGroup.IsGroup);

            int currentLayerIDinLayerGroup = -1;

            int firstLayerIDinLayerGroup = -1;
            LayerCollection layerCollection = layerGroup.ChildLayers;
            foreach(Layer layer in layerCollection)
            {
                if(layer.IsGroup)
                    continue;

                // ドキュメントのレイヤがあったらそのIDを返す
                if(layer.ID == documentCurrentLayerID)
                    return documentCurrentLayerID;

                if(firstLayerIDinLayerGroup == -1)
                    firstLayerIDinLayerGroup = layer.ID;
            }

            // レイヤグループ内の最初のレイヤのIDを返す
            if(firstLayerIDinLayerGroup >= 0)
                currentLayerIDinLayerGroup = firstLayerIDinLayerGroup;

            return currentLayerIDinLayerGroup;
        }

        /// <summary>
        /// レイヤグループの説明文字列から指定した項目名のIDを取得
        /// </summary>
        /// <param name="layerGroup"></param>
        /// <param name="strItemName"></param>
        /// <param name="findFirstIndex">説明分の文字列中に見つかった項目名の最初の文字のインデックス</param>
        /// <param name="findTextLength">説明分の文字列中に見つかった項目名の最初の文字から最後の文字までの文字数</param>
        /// <returns></returns>
        static public int GetIdFromLayerGroupDescription(Layer layerGroup, string strItemName, out int findFirstIndex, out int findTextLength)
        {
            findFirstIndex = -1;
            findTextLength = -1;

            int id = -1;
            if(layerGroup == null)
                return id;

            string searchText = strItemName;

            // レイヤグループの説明文字から
            string description = layerGroup.Description;

            // 仕様 ID部分の数字は()で囲まれている 例えばIDが2なら"RCAddInData_CurrentLayerID(2)"
            searchText += "(";

            int findIndex = description.IndexOf(searchText);
            if(findIndex == -1)
                return id;

            int searchTextLength = searchText.Length;

            int findIndex2 = description.IndexOf(")", findIndex + searchTextLength);
            if(findIndex2 == -1)
                return id;

            int isTextLength = findIndex2 - (findIndex + searchTextLength);
            string idText = description.Substring(findIndex + searchTextLength, isTextLength);

            //.文字から数値に変換
            int l;
            if(int.TryParse(idText, out l))
            {
                id = l;
                findFirstIndex = findIndex;// 説明文中に見つかった項目名の文字の最初のインデックス
                findTextLength = findIndex2 - findIndex + 1;// 項目とIDの書かれた文字の文字数
            }

            return id;            
        }

        /// <summary>
        /// レイヤグループの説明文字列に指定した項目名のIDをセット
        /// </summary>
        /// <param name="layerGroup"></param>
        /// <param name="strItemName"></param>
        /// <param name="id"> -1:削除 </param>
        static public void SetIdToLayerGroupDescription(Layer layerGroup, string strItemName, long id)
        {
            Debug.Assert(layerGroup != null);

            // レイヤグループの説明文字
            string description = layerGroup.Description;

            // レイヤグループの説明からカレントレイヤID取得
            int findFirstIndex;
            int findTextLength;
            int getID = GetIdFromLayerGroupDescription(layerGroup, strItemName, out findFirstIndex, out findTextLength);
            // 説明文中から取得でなかった場合やid が異なれば消す
            if(getID < 0 || getID != id)
            {
                if(findFirstIndex == 0)
                {
                    description = description.Remove(findFirstIndex, findTextLength);
                    // 先頭のカンマとスペースは削除
                    description = description.TrimStart(',', ' ');
                }
                else if(findFirstIndex > 0)
                {
                    int descriptionLength = description.Length;
                    string strLeft = description.Substring(0, findFirstIndex);
                    string strRight = description.Substring(findFirstIndex + findTextLength);

                    // 末尾の空白削除
                    string strLeft2 = strLeft.TrimEnd();
                    // 最後にカンマがあればそれも削除
                    int strLeft2Length = strLeft2.Length;
                    if(strLeft2Length > 0)
                    {
                        string strLeft3 = strLeft2.Substring(strLeft2Length - 1);
                        if(strLeft3 == ",")
                            strLeft = strLeft2.Substring(0, strLeft2Length - 1);
                    }

                    description = strLeft + strRight;
                }

                // 削除するだけの場合はここまで
                if(id == -1)
                {
                    // 説明セット
                    layerGroup.Description = description;
                    return;
                }
            }

            // 新しくセット
            if(getID != id)
            {
                // 仕様 ID部分の数字は()で囲まれている 例えばIDが2なら"RCAddInData_CurrentLayerID(2)"
                string setItemNameAndValue = strItemName + "(";

                string strValue = id.ToString();

                setItemNameAndValue += strValue + ")";

                // 新しくカンマで区切って追加(他に文字がない場合はカンマなし)
                if(description != "")
                    description += ", " + setItemNameAndValue;
                else
                    description += setItemNameAndValue;

                // 説明セット
                layerGroup.Description = description;
            }
        }
    }
}