﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Drawing;
using System.Diagnostics;
using System.IO;

using RootPro.RootProCAD;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.UI;

/// <summary locale="ja">
/// レイヤグループとスケール連携 プログラム
/// 概要
/// 　このプログラムは、レイヤグループと部分図レイアウトを連携させることができる RootPro CAD のアドインプログラムです。
/// 　RootPro CAD のスケールは部分図レイアウトに設定しますが、このアドインで、レイヤグループごとに部分図レイアウトを連携させると、
/// 　レイヤグループごとにスケールを設定しているような使い方ができます。
/// 　レイヤグループにはカレントレイヤも保持されるため、レイヤグループを切り替えたとき前回そのレイヤグループでカレントだったレイヤがカレントになります。
/// 　
/// メモ
///   ・レイヤグループの「説明」プロパティに連携させる部分図レイアウトやカレントレイヤのIDを書き込むことでアドインの動作を実現させています。
///   ・図面中には用紙は1枚、レイヤグループは16グループまでで使えます。
/// 
/// ※使用許諾
/// ・このアドインを実行した結果についての一切の責任はお使いになられる方にあります。
/// ・このアドインは、プログラムソース自体を改変・流用することが可能です。
/// 　また、このプログラムソースを元に独自のアドインやソフトウェアを作成し、無償・有償を問わず第三者に配布することも認めます。
/// 　ただし、改変・流用する場合は、株式会社ルートプロ(RootPro Corporation)作成の
/// 　当プログラムソースを参考にしていることを第三者にわかるように明記してください。
/// 　
/// 　Copyright @ 2022 RootPro Co.,Ltd. (Japan)
/// </summary>
/// <summary locale = "en">
/// Layer group and scale cooperation program
/// Overview
///  This program is a RootPro CAD add-in program that can link layer groups and part drawing layouts.
///  The scale of RootPro CAD is set to the part drawing layout, but if you link the part drawing layout for each layer group with this add-in,
///  It can be used as if the scale is set for each layer group.
///  Since the current layer is also retained in the layer group, the layer that was the current in that layer group last time when the layer group is switched becomes the current.
/// 
/// Note
///  It is realized by writing the part drawing layout and the ID of the current layer to be linked to the "Description" property of the layer group.
///  You can use up to 1 sheet of paper and up to 16 layer groups in the document.
/// 
/// License
///  The user is solely responsible for the result of executing this add-in.
///  This add-in can modify or divert the program source itself.
///  We also allow you to create your own add-ins and software based on this program source and distribute them to third parties regardless of whether they are free or paid.
///  However, if you want to modify or divert it, create it by RootPro Corporation.
///  Please specify so that a third party can understand that you are referring to this program source.
/// 
///  Copyright @ 2022 RootPro Co., Ltd. (Japan)
/// </summary>

namespace RCAddInSelectLayerGroupCS
{
    partial class RootProCADUserControl:UserControl
    {
        AppAddIn app;

        // 説明図面のファイル名
        string helpFileName = "RcAddInSelectLayerGroupManual.rpcd";// アドインのdllと同じ場所

        private bool useRootProCADDisplayFont = false;// フォーム上のフォントは（サイズ）も RootPro CAD と同じにするフラグ（Load 時に値を取得するのみ）

        public const int RcAddInLayerGroupSize = 16;// レイヤグループの数
        public const int RcAddInMaxLayerGroupSize = RcAddInLayerGroupSize;// レイヤグループの数の最大値
        public const int RcAddInChildLayerSize = 16;// レイヤグループ内のレイヤの数
        public const int RcAddInMaxChildLayerSize = RcAddInChildLayerSize;// レイヤグループ内のレイヤの数の最大値

        public const string ItemName_CurrentLayerID = "RC_LayerID";// レイヤグループ内のカレントレイヤIDを表す名前（レイヤグループの説明に文字列で保持）
        public const string ItemName_InterlockingDrawingLayoutID = "RC_LayoutID";// レイヤグループに連動させる部分図レイアウトIDを表す名前（レイヤグループの説明に文字列で保持）

        // カレントレイヤ変更イベント処理中フラグ
        bool isAppCurrentLayerChanging = false;
        // カレント部分図レイアウト変更イベント処理中フラグ
        bool isAppCurrentDrawingLayoutChanging = false;

        // コントロール Refresh 有効フラグ
        bool refreshControl = true;

        // 選んでいたタブのインデックス
        private int lastSelectTabIndex = -1;
        // ツールチップが表示されたときのタブのインデックス
        private int lastToolTipSelectTabIndex = -1;

        // コンテキストメニューが開くとき選んだタブのインデックス
        int contextMenuSelectTabIndex = -1;

        bool changeDisplayFont = false;// 画面の表示フォントを変更するかどうかを表す値
        string fontName = "";// 画面の表示フォント名
        double fontRatio = 1.0;// フォントの表示倍率

        private bool setEventHandler = false;// イベントハンドラをセットしたかどうかを表す値

        public static int GetRcAddInMaxChildLayerSize()
        {
            return RcAddInMaxChildLayerSize;
        }

        // 数値を比較するときに使用する精度
        double RcPrecisionConfusion = Precision.Confusion;
        double RcPrecisionAngular = Precision.RadianAngular;

        // レイヤグループ＆レイヤコントロール
        public LayerGroupsLayerControl[] layerGroupsLayerControls;

        public RootProCADUserControl(AppAddIn app)
        {
            this.app = app;
            InitializeComponent();


            AddInManager addInManager = app.Application.AddInManager;
            useRootProCADDisplayFont = addInManager.UseDisplayFont && app.addInSettings.rootProCADDisplayFont ? true:false;// フォーム上のフォントは（サイズ）もRootPro CAD と同じにするフラグ

            this.layerGroupsLayerControls = null;

            this.Load += RootProCADUserControl_Load;
        }

        // このドッキングウィンドウの表示開始時処理
        private void RootProCADUserControl_Load(object sender, EventArgs e)
        {
            // レイヤグループ＆レイヤコントロール作成とタブに追加
            CreateLayerGroupsLayerControl();
            
            ApplicationSettings appSettings = app.Settings;
            this.changeDisplayFont = appSettings.ChangeDisplayFont;// 画面の表示フォントを変更するかどうかを表す値
            this.fontName = appSettings.DisplayFontName;// 画面の表示フォント名
            this.fontRatio = appSettings.DisplayFontRatio;// フォントの表示倍率

            // ツールチップ
            string tipText = Properties.Resources.groupToolTipLayerGroupTabControl;
            groupToolTip.SetToolTip(layerGroupTabControl, tipText);

            // 各コントロールの状態更新
            UpdateFormControlState();

            // イベントハンドラーセット
            SetEventHandler();            
        }

        /// <summary>
        /// RootPro CAD の［オプション］が変更したときの処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void App_SettingsChanged(object sender, ApplicationSettingsEventArgs e)
        {
            bool adjustFloatingBar = false;// 設定の変更前と値が違えばフローティングバーのサイズを合わせるフラグ
            ApplicationSettings appSettings = e.Settings;
            bool newChangeDisplayFont = appSettings.ChangeDisplayFont;// 画面の表示フォントを変更するかどうかを表す値
            string newFontName = appSettings.DisplayFontName;// 画面の表示フォント名
            double newFontRatio = appSettings.DisplayFontRatio;// フォントの表示倍率
            Debug.Assert(app.addInSettings != null);
            if(app.Application.AddInManager.UseDisplayFont && app.addInSettings.rootProCADDisplayFont && (this.changeDisplayFont != newChangeDisplayFont || this.fontName != newFontName || this.fontRatio != newFontRatio))
            {
                // フローティングバーのサイズを合わせる
                adjustFloatingBar = true;
            }

            // コントロールのサイズ更新
            UpdateControlsSize();

            if(adjustFloatingBar)
                AdjustFloatingBar();
           
        }

        /// <summary>
        /// ドッキングウィンドウがフローティング状態になったときのイベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AppAddIn_DockingBarFloated(object sender, DockingBarEventArgs e)
        {
            DockingBar dockingBar = e.DockingBar;
            if(dockingBar is CustomDockingBar)
            {
                CustomDockingBar customDockingBar = (CustomDockingBar)dockingBar;
                if(customDockingBar.UniqueName == app.myAddInUniqueName)
                {
                    // Adjust FloatingBar
                    AdjustFloatingBar();
                }
            }
        }

        /// <summary>
        /// タブコントロールの各ページにレイヤグループ＆レイヤコントロールを追加
        /// </summary>
        private void CreateLayerGroupsLayerControl()
        {
            // フォームのレイアウトによる再描画一時停止
            this.SuspendLayout();

            TabControl.TabPageCollection tabPages = layerGroupTabControl.TabPages;
            int tabCount = tabPages.Count;
            Debug.Assert(tabCount > 0);
            this.layerGroupsLayerControls = new LayerGroupsLayerControl[tabCount];

            for(int i = 0; i < tabCount; i++)
            {
                layerGroupsLayerControls[i] = new LayerGroupsLayerControl(this);

                // コントロールのプロパティ
                layerGroupsLayerControls[i].Name = "layerGroupsLayerControl" + i.ToString();
                layerGroupsLayerControls[i].Location = new Point(0, 0);

                // インデックスセット
                layerGroupsLayerControls[i].layerGroupsLayerControlIndex = i;

                // コントロールをクリア状態
                layerGroupsLayerControls[i].ClearControls();

                // 各ページへ追加
                tabPages[i].Controls.Add(layerGroupsLayerControls[i]);
            }
  
            // フォームのレイアウトによる再描画処理再開
            this.ResumeLayout();

            // コントロールのサイズ更新
            UpdateControlsSize(true);
        }


        /// <summary>
        /// コントロールのサイズ更新
        /// </summary>
        private void UpdateControlsSize(bool createControl = false)
        {
            TabControl.TabPageCollection tabPages = layerGroupTabControl.TabPages;
            int tabCount = tabPages.Count;
            if(tabCount < 1)
                return;
            
            if(layerGroupsLayerControls.Length < 1)
                return;

            // フォームのレイアウトによる再描画一時停止
            this.SuspendLayout();

            // タブの部分のサイズ変更
            int itemSizeWidth = 34;
            int itemSizeHeight = 72;// (タブは縦に並べているのでSizeのHeightが横方向)
            ApplicationSettings appSettings = app.Application.Settings;// フォントの表示倍率
            double fontRatio = useRootProCADDisplayFont ? appSettings.DisplayFontRatio:1.0;
            layerGroupTabControl.ItemSize = new Size((int)(itemSizeWidth * fontRatio), (int)(itemSizeHeight * fontRatio));

            // 各ページのコントロールについて更新する
            for(int i = 0; i < tabCount; i++)
                layerGroupsLayerControls[i].AlignmentAndVisibleControls();

            // タブコントロールのサイズ更新
            // 最初のページのレイヤボックスの幅を取得して、それに合わせる
            Rectangle layerGroupsLayerControlClientRect = layerGroupsLayerControls[0].Bounds;
            double sizeRatio = createControl ? 1.0 / fontRatio:1.0;// 初回だけsizeRatioで調整が必要(メモ：RootPro CAD にドッキングウィンドウの表示イベントがあればいいと思う。_VisibleChanged イベントでは期待する値が取得できなかった)
            int layerGroupTabControlSizeWidth = (int)((layerGroupTabControl.ItemSize.Height * sizeRatio + layerGroupsLayerControlClientRect.Width));
            int layerGroupTabControlSizeHeight = layerGroupsLayerControlClientRect.Height;
            int marginX = layerGroupTabControl.Margin.Left;
            int marginY = layerGroupTabControl.Margin.Top;
            int paddingX = layerGroupTabControl.Padding.X; 
            int paddingY = layerGroupTabControl.Padding.Y;
            layerGroupTabControl.Size = new Size(layerGroupTabControlSizeWidth + marginX * 2 + paddingX * 2, layerGroupTabControlSizeHeight + marginY * 2 + paddingY * 2);

            // 「レイヤグループと部分図」枠(グループボックス)のサイズ更新
            int marginX2 = layerGroupTabControl.Location.X;
            int marginY2 = layerGroupTabControl.Location.Y;
            layerGroup_GroupBox.Size = new Size(layerGroupTabControl.Size.Width + marginX2 * 2, layerGroupTabControl.Size.Height + marginY2);
            // コンパクトモードのときの文字
            if(IsCompactFormMode())
                layerGroup_GroupBox.Text = Properties.Resources.layerGroup_GroupBox_Text_Compact;
            else
                layerGroup_GroupBox.Text = Properties.Resources.layerGroup_GroupBox_Text_Normal;


            // メンバ更新
            this.changeDisplayFont = appSettings.ChangeDisplayFont;// 画面の表示フォントを変更するかどうかを表す値
            this.fontName = appSettings.DisplayFontName;// 画面の表示フォント名
            this.fontRatio = appSettings.DisplayFontRatio;// フォントの表示倍率

            // フォームのレイアウトによる再描画処理再開
            this.ResumeLayout();
        }

        /// <summary>
        /// フォームのフォント変更時のイベント処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RootProCADUserControl_FontChanged(object sender, EventArgs e)
        {
            // コンテキストメニューのフォント
            // フォーム上に配置されているコントロールのフォントと同じにする
            groupContextMenuStrip.Font = layerGroup_GroupBox.Font;

            ApplicationSettings appSetting = app.Application.Settings;// フォントの表示倍率
            double fontRatio = useRootProCADDisplayFont ? appSetting.DisplayFontRatio:1.0;

            // コンテキストメニューのアイコンサイズ
            int defImageSize = 24;
            double imageScale = 1.0 + (fontRatio - 1.0) * 0.8;
            groupContextMenuStrip.ImageScalingSize = new Size((int)(defImageSize * imageScale), (int)(defImageSize * imageScale));
        }

        /// <summary>
        /// ハンドルの破棄イベント処理
        /// RootPro CAD 終了時のメッセージ処理としてこれを使う
        /// </summary>
        /// <param name="e"></param>
        protected override void OnHandleDestroyed(EventArgs e)
        {

            // 設定値をファイルに書き込む
            Debug.Assert(app.addInSettings != null);
            //app.GetAddInSettings(app.addInSettings);
            app.SaveAddInSettings(app.addInSettings);

            RemoveEventHandler();

            RemoveAppCommand();

            // 基底クラスのイベント
            base.OnHandleDestroyed(e);
        }

        private void RemoveAppCommand()
        {
        }


        private void SetEventHandler()
        {
            if (setEventHandler)
                return;

            // 新しい図面が作成さたときのイベントハンドラ
            app.NewDocument += new DocumentEventHandler(App_NewDocument);
            // 図面を開いたときのイベントハンドラ追加
            app.DocumentOpened += new DocumentEventHandler(App_DocumentOpened);
            // アクティブな図面変更イベントハンドラ追加
            app.DocumentActivated += new DocumentEventHandler(App_DocumentActivated);
            // 図面が保存される前のイベントハンドラ追加
            app.DocumentBeforeSave += new DocumentBeforeSaveEventHandler(App_DocumentBeforeSave);

            // カレント部分図変更イベントハンドラ追加
            app.CurrentDrawingChanged += new CurrentDrawingChangedEventHandler(App_CurrentDrawingChanged);
            // 部分図追加イベントハンドラ追加
            app.DrawingAdded += new DrawingEventHandler(App_DrawingAdded);
            // 部分図変更イベントハンドラ追加
            app.DrawingChanged += new DrawingChangedEventHandler(App_DrawingChanged);

            // カレント部分図レイアウトの変更イベントハンドラ追加（このイベントは取れないが部分図の変更イベントが起きるのでそこでカレントの部分図レイアウトが変わったか分かる）
            // 部分図レイアウト追加イベントハンドラ追加
            app.LayoutAdded += new RootPro.RootProCAD.LayoutEventHandler(App_LayoutAdded);
            // 部分図レイアウト変更イベントハンドラ追加
            app.LayoutChanged += new LayoutChangedEventHandler(App_LayoutChanged);
            // 部分図レイアウト削除イベントハンドラ追加
            app.LayoutDeleting += new LayoutDeletingEventHandler(App_LayoutDeleting);
         
            // カレントレイヤ変更イベントハンドラ追加
            app.CurrentLayerChanged += new CurrentLayerChangedEventHandler(App_CurrentLayerChanged);
            // レイヤ追加イベントハンドラ追加
            app.LayerAdded += new LayerEventHandler(App_LayerAdded);
            // レイヤ変更イベントハンドラ追加
            app.LayerChanged += new LayerChangedEventHandler(App_LayerChanged);
            // レイヤの表示順変更イベントハンドラ追加
            app.LayerOrderChanged += new LayerOrderChangedEventHandler(App_LayerOrderChanged);
            // レイヤ削除イベントハンドラ追加
            app.LayerDeleting += new LayerDeletingEventHandler(App_LayerDeleting);

            // 図形に関するイベントハンドラ追加
            app.ShapeAdded += new ShapeEventHandler(App_ShapeAdded);
            app.ShapeDelete += new ShapeEventHandler(App_ShapeDelete);
            app.ShapeChanged += new ShapeEventHandler(App_ShapeChanged);

            // タブビュー関連のイベントハンドラ
            app.TabViewActivated += new TabViewEventHandler(App_TabViewActivated);

            // オプション設定が変更されたとのイベントハンドラ
            app.ApplicationSettingsChanged += new ApplicationSettingsEventHandler(App_SettingsChanged);

            // ドッキングウィンドウがフローティング状態になったときのイベントハンドラ
            app.DockingBarFloated += new DockingBarEventHandler(AppAddIn_DockingBarFloated);

            setEventHandler = true;
        }

        private void RemoveEventHandler()
        {
            if (!setEventHandler)
                return;

            // 新しい図面が作成さたときのイベントハンドラ
            app.NewDocument -= new DocumentEventHandler(App_NewDocument);
            // 図面を開いたときのイベントハンドラ削除
            app.DocumentOpened -= new DocumentEventHandler(App_DocumentOpened);
            // アクティブな図面変更イベントハンドラ削除
            app.DocumentActivated -= new DocumentEventHandler(App_DocumentActivated);
            // 図面が保存される前のイベントハンドラ削除
            app.DocumentBeforeSave -= new DocumentBeforeSaveEventHandler(App_DocumentBeforeSave);

            // カレント部分図変更イベントハンドラ削除
            app.CurrentDrawingChanged -= new CurrentDrawingChangedEventHandler(App_CurrentDrawingChanged);
            // 部分図追加イベントハンドラ削除
            app.DrawingAdded -= new DrawingEventHandler(App_DrawingAdded);
            // 部分図変更イベントハンドラ削除
            app.DrawingChanged -= new DrawingChangedEventHandler(App_DrawingChanged);
         
            // 部分図レイアウト追加イベントハンドラ削除
            app.LayoutAdded -= new RootPro.RootProCAD.LayoutEventHandler(App_LayoutAdded);
            // 部分図レイアウト変更イベントハンドラ削除
            app.LayoutChanged -= new LayoutChangedEventHandler(App_LayoutChanged);
            // 部分図レイアウト削除イベントハンドラ削除
            app.LayoutDeleting -= new LayoutDeletingEventHandler(App_LayoutDeleting);

            // カレントレイヤ変更イベントハンドラ削除
            app.CurrentLayerChanged -= new CurrentLayerChangedEventHandler(App_CurrentLayerChanged);
            // レイヤ追加イベントハンドラ削除
            app.LayerAdded -= new LayerEventHandler(App_LayerAdded);
            // レイヤ変更イベントハンドラ削除
            app.LayerChanged -= new LayerChangedEventHandler(App_LayerChanged);
            // レイヤの表示順変更イベントハンドラ削除
            app.LayerOrderChanged -= new LayerOrderChangedEventHandler(App_LayerOrderChanged);
            // レイヤ削除イベントハンドラ削除
            app.LayerDeleting -= new LayerDeletingEventHandler(App_LayerDeleting);

            // 図形に関するイベントハンドラ削除
            app.ShapeAdded -= new ShapeEventHandler(App_ShapeAdded);
            app.ShapeDelete -= new ShapeEventHandler(App_ShapeDelete);
            app.ShapeChanged -= new ShapeEventHandler(App_ShapeChanged);

            // タブビュー関連のイベントハンドラ
            app.TabViewActivated -= new TabViewEventHandler(App_TabViewActivated);

            // オプション設定が変更されたとのイベントハンドラ
            app.ApplicationSettingsChanged -= new ApplicationSettingsEventHandler(App_SettingsChanged);

            // ドッキングウィンドウがフローティング状態になったときのイベントハンドラ
            app.DockingBarFloated -= new DockingBarEventHandler(AppAddIn_DockingBarFloated);

            setEventHandler = false;
        }
         
        /// <summary>
        /// 新しい図面が作成さたときのイベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_NewDocument(object sender, DocumentEventArgs e)
        {
            UpdateFormControlState();
        }

        /// <summary>
        /// 図面を開いたときのイベントハンドラ追加
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_DocumentOpened(object sender, DocumentEventArgs e)
        {
            Document doc = e.Document;

            string documentName = doc.Name;
            int documentNameLength = documentName.Length;

            // RootPro CAD 以外のファイルを読み込んで拡張子がJwwだった場合
            string jwwDotExtension = ".jww";
            int jwwDotExtensionLength = jwwDotExtension.Length;
            if(documentNameLength > jwwDotExtensionLength)
            {
                string documentNameDotExtension = documentName.Substring(documentNameLength - jwwDotExtensionLength, jwwDotExtensionLength);
                if(String.Compare(documentNameDotExtension, jwwDotExtension, true) == 0)
                {
                    // "Jw_cad のファイルを読み込みました。\nレイヤグループにあったスケールは RootPro CAD では部分図レイアウトに設定されています。\nこのアドインを使うとカレントレイヤのあるレイヤグループとそのスケールが設定された部分図レイアウトを関連付けします。関連付けを実行しますか?\n関連付けが可能な場合はレイヤグループのプロパティの「説明」に関連付けした部分図レイアウト情報が書かれます(このレイヤグループの説明に書かれた部分図レイアウト情報は直接編集しないでください)。"
                    string msg = Properties.Resources.documentOpenedMsg_1;
                    if(DialogResult.Yes == MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Information))
                    {
                        // レイヤグループと部分図レイアウトの連動できる状態ならレイヤグループの説明に部分図レイアウト情報を書く
                        if(!IsConvertableDocument(doc, out string strErr))
                        {
                            //"レイヤグループと部分図レイアウトの関連付けができませんでした。\n\n" + strErr;
                            msg = Properties.Resources.documentOpenedMsg_2 + "\n\n" + strErr;
                            MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }
                        else
                        {
                            // レイヤグループに部分図レイアウトを関連付け実行(Undo 履歴は取らない(Undo によって説明プロパティに書き込んだ情報が消えないようにするため))
                            bool result = InterlockingDrawinglayoutToLayerGroup(doc, false);
                            if(!result)
                            {
                                //"レイヤグループと部分図レイアウトの関連付けができませんでした。";
                                msg = Properties.Resources.documentOpenedMsg_3;
                                MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
                            }
                            else
                            {
                                // すべての部分図を表示のみにする(用紙と非表示の部分図は除く)。
                                // 部分図タブビューをすべて閉じる。
                                // 「[表示のみ]のレイヤ、部分図上の図形を選択する」を Off 推奨設定
                                // これらの状態は、部分図をまたがったレイヤ間の図形のコピー貼り付けを無意識にしてしまうことを防げる。

                                Cursor preCursor = Cursor.Current;
                                Cursor.Current = Cursors.WaitCursor;// 待機カーソルに

                                this.refreshControl = false;// コントロール Refresh 有効フラグ off

                                // 用紙タブビューをアクティブにする
                                ActiveFirstPaperTabView();

                                // 全ての編集可能な部分図を表示のみにする
                                SetAllPartDrawingStateEditableToVisible();

                                // すべての部分図タブを閉じる
                                CloseAllPartDrawingView();

                                this.refreshControl = true; // コントロール Refresh 有効フラグ On

                                Cursor.Current = preCursor;// カーソルを元に戻す

                                //"レイヤグループと部分図レイアウトの関連付けができました。";
                                msg = Properties.Resources.documentOpenedMsg_4;
                                msg += "\n\n";
                                msg += Properties.Resources.documentOpenedMsg_5;
                                MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Information);

                                // ドキュメントのカレントレイヤのあるタブのレイヤーグループに連携している部分図レイアウトをカレントにする
                                SetCurrentInterlockingDrawinglayoutInLayerGroupIncludeCurrentLayer();

                                // 「[表示のみ]のレイヤ、部分図上の図形を選択する」を Off 推奨設定
                                RecommendedCannotSelectShapeInShowOnlyLayerAndPartDrawing();

                                // ドッキングウィンドウを表示する
                                CustomDockingBar myDockingBar = app.GetMyDockingBar();
                                if (!myDockingBar.Visible)
                                    myDockingBar.Activate();
                            }
                        }

                        // UndoUnit 履歴クリア
                        doc.UndoManager.Clear();
                    }
                }
            }

            UpdateFormControlState();
        }

        /// <summary>
        /// アクティブなドキュメント変更イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_DocumentActivated(object sender, DocumentEventArgs e)
        {
            // 図面が変わるのでここで layerGroupTabControl の全てのタブで ClearControlsを 実行
            LayerGroupTabControlAllClear();

            UpdateFormControlState();
        }

        /// <summary>
        /// layerGroupsLayerControls 全てのタブで ClearControls を実行
        /// </summary>
        private void LayerGroupTabControlAllClear()
        {
            TabControl.TabPageCollection tabPages = layerGroupTabControl.TabPages;
            int tabCount = tabPages.Count;
            if(tabCount < 1)
                return;

            // コントロールをクリア
            for(int i = 0; i < tabCount; i++)
                layerGroupsLayerControls[i].ClearControls();
        }

        /// <summary>
        /// コントロールを有効無効にする
        /// </summary>
        /// <param name="enabled"></param>
        private void SetControlEnabled(bool enabled)
        {
            // グループボックス
            layerGroup_GroupBox.Enabled = enabled;

            // レイヤーグループタブコントロール
            layerGroupTabControl.Enabled = enabled;
        }

        // メモ：図面が切り替わろうとしたとき、今のタブのカレントレイヤ情報は
        // レイヤグループの説明に保存されないが、再度今のドキュメントに切り替わって
        // 戻った時にUpdateFormControlStateでタブの状態の整合性は保たれる。
        
        /// <summary>
        /// 図面が保存される前のイベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_DocumentBeforeSave(object sender, DocumentBeforeSaveEventArgs e)
        {
            Document doc = e.Document;

            // 今のタブのカレントレイヤIDをレイヤの説明に保存
            int tabIndex = layerGroupTabControl.SelectedIndex;
            UpdateLayerGroupDescription(doc, tabIndex);
        }

        /// <summary>
        /// カレント部分図の変更イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_CurrentDrawingChanged(object sender, CurrentDrawingChangedEventArgs e)
        {
            // ※ カレント部分図レイアウトの変更イベント（このイベントは取れないが、カレント部分図の変更イベントが起きるのでそこでカレントの部分図レイアウトが変わったか分かる)

            Document doc = app.ActiveDocument;
            Debug.Assert(doc != null);
            if(doc == null)
                return;

            // カレント部分図レイアウト変更イベント処理中フラグ
            isAppCurrentDrawingLayoutChanging = true;

            // フォームの状態更新
            UpdateFormControlState(null, true);

            // タブが切り替わった場合、カレントレイヤもそのレイヤグループのものに変える
            // レイヤグループの説明からカレントレイヤIDを取得
            int tabIndex = layerGroupTabControl.SelectedIndex;
            Layer layerGroup = layerGroupsLayerControls[tabIndex].GetLayerGroup();
            if(layerGroup == null)
                return;

            int currentLayerID = GetIdFromLayerGroupDescription(layerGroup, ItemName_CurrentLayerID, out int damyIndex1, out int damyIndex2);
            // IDが有効かチェック
            if(!AddInTools.IsValidLayerID(doc, currentLayerID))
                currentLayerID = -1;

            if(currentLayerID >= 0)
            {
                LayerTable layerTable = doc.LayerTable;
                Layer layer = layerTable.GetLayerByID(currentLayerID);

                // グループのカレントレイヤをドキュメントにセット(ドキュメントのカレントレイヤと異なれば)
                if(layer != null && !layerGroupsLayerControls[tabIndex].IsDocumentCurrentLayer())
                    layerTable.CurrentLayer = layer;
            }
	    
            isAppCurrentDrawingLayoutChanging = false;
        }

        /// <summary>
        /// RootPro CAD のタブビューがアクティブになったときのイベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_TabViewActivated(object sender, TabViewEventArgs e)
        {
            TabView tabView = e.TabView;
            if(tabView == null)
                return;

            // 部分図タブビューの場合
            bool drawingChanged = false;
            if(tabView.TabViewType == TabViewType.PartView)
                drawingChanged = true;

            UpdateFormControlState(null, drawingChanged);
        }

        /// <summary>
        /// 部分図追加イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_DrawingAdded(object sender, DrawingEventArgs e)
        {
            UpdateFormControlState(null, true);
        }

        /// <summary>
        /// 部分図変更イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_DrawingChanged(object sender, DrawingChangedEventArgs e)
        {
            UpdateFormControlState(null, true);
        }

        // 部分図レイアウト追加イベントハンドラ追加
        void App_LayoutAdded(object sender, RootPro.RootProCAD.LayoutEventArgs e)
        {
            // コントロール Refresh 有効フラグ off
            bool oldRefreshControl = this.refreshControl;
            this.refreshControl = false;

            UpdateFormControlState(null, true);

            // コントロール Refresh 有効フラグを戻す
            this.refreshControl = oldRefreshControl;
        }

        // 部分図レイアウト変更イベントハンドラ追加
        void App_LayoutChanged(object sender, LayoutChangedEventArgs e)
        {
            UpdateFormControlState(null, true);
        }

        // 部分図レイアウト削除イベントハンドラ追加
        void App_LayoutDeleting(object sender, LayoutDeletingEventArgs e)
        {
            Layout layout = e.Layout;
            UpdateFormControlState(null, true, layout.ID);
        }

        /// <summary>
        /// カレントレイヤの変更イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_CurrentLayerChanged(object sender, CurrentLayerChangedEventArgs e)
        {
            // カレントレイヤ変更イベント処理中フラグ
            this.isAppCurrentLayerChanging = true;

            // フォームの状態更新
            UpdateFormControlState(null);

            this.isAppCurrentLayerChanging = false;
        }
        
        void App_LayerAdded(object sender, LayerEventArgs e)
        {
            UpdateFormControlState();
        }

        void App_LayerChanged(object sender, LayerChangedEventArgs e)
        {
            Layer layer = e.Layer;
            UpdateFormControlState(layer);
        }

        // レイヤの表示順が変わったとき
        void App_LayerOrderChanged(object sender, LayerOrderChangedEventArgs e)
        {
            UpdateFormControlState();
        }

        void App_LayerDeleting(object sender, LayerDeletingEventArgs e)
        {
            Layer layer = e.Layer;
            UpdateFormControlState(null, false, layer.ID);
        }

        /// <summary>
        /// 図形が追加されたとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_ShapeAdded(object sender, ShapeEventArgs e)
        {
            // 図形のレイヤのあるレイヤグループのタブとそのページのレイヤボタンを再描画
            InvalidateLayerTabAndButtonControl(e.Shape.LayerID);
        }

        /// <summary>
        /// 図形が削除されたとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_ShapeDelete(object sender, ShapeEventArgs e)
        {
            // 図形のレイヤのあるレイヤグループのタブとそのページのレイヤボタンを再描画
            InvalidateLayerTabAndButtonControl(e.Shape.LayerID);
        }

        /// <summary>
        /// 図形に変更があったとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_ShapeChanged(object sender, ShapeEventArgs e)
        {
            // 図形のレイヤIDが変わった場合
            // 変更前の図形を取得するためにShapeChangedEventArgsでキャスト
            if(e.Shape.LayerID != ((ShapeChangedEventArgs)e).OldShape.LayerID)
            { 
                // 図形のレイヤのあるレイヤグループのタブとそのページのレイヤボタンを再描画
                // 変更前のレイヤに対して
                InvalidateLayerTabAndButtonControl(((ShapeChangedEventArgs)e).OldShape.LayerID);
                // 変更後のレイヤに対して
                InvalidateLayerTabAndButtonControl(e.Shape.LayerID);
            }
        }

        /// <summary>
        /// 指定したレイヤのあるレイヤグループのタブとボタンを再描画
        /// </summary>
        /// <param name="layerID"></param>
        void InvalidateLayerTabAndButtonControl(int layerID)
        {
            // 指定したレイアIDのレイアがあるタブを探す
            int findedTabIndex = FindTabIndexIncludeLayerId(layerID);
            if(findedTabIndex >= 0)
            {
                // タブ部分の再描画
                Rectangle tabItemRect = layerGroupTabControl.GetTabRect(findedTabIndex);
                layerGroupTabControl.Invalidate(tabItemRect);

                // レイヤの状態ボタン再描画
                LayerStateButton layerStateButton = layerGroupsLayerControls[findedTabIndex].GetLayerStateButton(layerID);
                if(layerStateButton != null)
                    layerStateButton.Invalidate();
            }
        }

        /// <summary>
        /// フォームの状態更新
        /// </summary>
        /// <param name="layer"></param>
        /// <param name="drawingChanged"></param>
        /// <param name="deleteingID">レイヤか部分図レイアウトの削除するときのそれぞれのID</param>
        private void UpdateFormControlState(Layer layer = null, bool drawingChanged = false, int deleteingID = -1)
        {
            Document doc = app.ActiveDocument;

            // 用紙の数とルート直下にレイヤがあるか
            int paperCount = 0;
            bool isRootLayer = false;
            if(doc != null)
            {
                paperCount = doc.Drawings.PaperCount;
                // ルートにレイヤがあるか
                if(drawingChanged)
                    isRootLayer = AddInTools.IsRootLayers(doc);
                // 削除するレイヤ以外にルートにレイヤがあるか
                else
                    isRootLayer = AddInTools.IsOtherRootLayers(doc, deleteingID);
            }

            // 用紙は1つで、ルート直下のレイヤはなく、作図編集ウィンドウで用紙や部分図以外のタブ（レイヤ一覧や作図部品編集）が選ばれているときの動作（コントロールをグレー表示など）
            AddInTools.RC_TabViewType tabViewType = AddInTools.GetActiveTabViewType(app);
            if(doc != null && paperCount == 1 && !isRootLayer && (tabViewType == AddInTools.RC_TabViewType.RC_TabViewType_Paper || tabViewType == AddInTools.RC_TabViewType.RC_TabViewType_Part))
                SetControlEnabled(true);
            else 
                SetControlEnabled(false);

            TabControl.TabPageCollection tabPages = layerGroupTabControl.TabPages;
            int tabCount = tabPages.Count;

            if(doc == null || layerGroupTabControl.Enabled == false)
            {
                 for(int i = 0; i < tabCount; i++)
                    layerGroupsLayerControls[i].SetLayerGroup(null, null);

                // コントロールの再描画処理
                RefreshControls();

                return;
            }

            // レイヤが指定されている場合
            if(layer != null)
            {
                // 親グループレイヤIDが一致するタブを探す
                int layerGroupID = -1;
                if(layer.IsGroup)
                    layerGroupID = layer.ID;
                else
                    layerGroupID = layer.ParentLayer.ID;
                
                if(layerGroupID != -1)
                {
                    for(int i = 0; i < tabCount; i++)
                    {
                        if(layerGroupsLayerControls[i].GetLayerGroupID() == layerGroupID)
                        {
                            // 状態再描画
                            layerGroupsLayerControls[i].UpdateLayerState(layer);
                            break;
                        }
                    }
                }
            }
            // レイヤが指定されていな場合
            else
            {
                // 部分図レイアウトではなく、レイヤ関係の更新
                if(!drawingChanged)
                {
                    int deleteingLayerID = deleteingID;
                    
                    // レイヤグループのリスト取得(16個まで)
                    List<Layer> layerGroupList = GetLayerGroup16List(doc, deleteingLayerID);

                    // 各タブコントロールにレイヤグループセット
                    int layerGroupListSize = 0;
                    if(layerGroupList != null)
                        layerGroupListSize = layerGroupList.Count;

                    bool selectedTab = false;
                    for(int i = 0; i < RcAddInMaxLayerGroupSize; i++)
                    {
                        if(i < layerGroupListSize)
                        {
                            Layer layerGroup = layerGroupList[i];
                            layerGroupsLayerControls[i].SetLayerGroup(doc, layerGroup, deleteingLayerID);

                            // レイヤグループ内のレイヤにドキュメントのカレントレイヤがある場合はそのタブを選択
                            if(!selectedTab && layerGroupTabControl.SelectedIndex != i && layerGroupsLayerControls[i].IsDocumentCurrentLayer())
                            {
                                layerGroupTabControl.SelectedIndex = i;
                                
                                if(layerGroup != null)
                                {
                                    // 関連している部分図レイアウトがあった場合
                                    int layoutID = GetIdFromLayerGroupDescription(layerGroup, ItemName_InterlockingDrawingLayoutID, out int findFirstIndex, out int findTextLength);
                                    Layout layout = AddInTools.GetLayoutOfID(doc, layoutID);
                                    if(layout != null)
                                    {
                                        // 作図編集ウィンドウのタブビューが用紙の場合
                                        // 関連している部分図レイアウトをカレントにする
                                        if(tabViewType == AddInTools.RC_TabViewType.RC_TabViewType_Paper)
                                        {
                                            // 部分図レイアウトをカレントにする
                                            if(!doc.CurrentLayout.Equals(layout))
                                                doc.CurrentLayout = layout;                                        
                                        }
                                        // 作図編集ウィンドウのタブビューが部分図の場合
                                        // 関連している部分図レイアウトの元の部分図のタブビューに切り替える
                                        else if(tabViewType == AddInTools.RC_TabViewType.RC_TabViewType_Part)
                                        {
                                            // 部分図をカレントにする
                                            if(!doc.CurrentDrawing.Equals(layout.Drawing))
                                                doc.CurrentDrawing = layout.Drawing;
                                        }
                                    }
                                }

                                selectedTab = true;
                            }
                        }
                        // レイヤグループがないタブ
                        else
                        {
                            layerGroupsLayerControls[i].SetLayerGroup(doc, null);
                        }
                    }
                }
                // 部分図レイアウト関係の更新処理
                else
                {
                    int deleteingLayoutID = deleteingID;
                    bool isInterlockingDrawingLayout = false;
                    int selectTabIndex = -1;

                    // 作図編集ウィンドウで用紙タブが選ばれている場合
                    if(AddInTools.GetActiveTabViewType(app) == AddInTools.RC_TabViewType.RC_TabViewType_Paper)
                    {
                        // カレントの部分図レイアウトが連動する部分図レイアウトとなっているレイヤグループがあればそのタブを選択する
                        int docCurrentLayoutID = -1;
                        Layout layout = doc.CurrentLayout;
                        if(layout != null)
                            docCurrentLayoutID = layout.ID;

                        if(docCurrentLayoutID >= 0)
                        {
                            // レイヤグループに関連づいた部分図レイアウトのリストのインデックス
                            int findIndex = FindInterlockingDrawingLayoutIdInLayerGroup(docCurrentLayoutID);
                            if(findIndex >= 0)
                            {
                                // 切り替えるタブのインデックス保持
                                isInterlockingDrawingLayout = true;
                                if(layerGroupTabControl.SelectedIndex != findIndex)
                                    selectTabIndex = findIndex;
                            }
                        }
                    }
                    else
                    {
                        // カレントの部分図が元となった部分図レイアウトが関連付けられているレイヤグループのタブを探してあったらそれを選ぶ
                        Drawing drawing = doc.CurrentDrawing;

                        if(drawing != null && drawing.Name != "")
                        {
                            for(int i = 0; i < RcAddInMaxLayerGroupSize; i++)
                            {
                                Layer layerGroup = layerGroupsLayerControls[i].GetLayerGroup();
                                if(layerGroup != null)
                                {
                                    int layoutID = GetIdFromLayerGroupDescription(layerGroup, ItemName_InterlockingDrawingLayoutID, out int findFirstIndex, out int findTextLength);
                                    Layout layout = AddInTools.GetLayoutOfID(doc, layoutID);
                                    if(layout != null)
                                    {
                                        // 元の部分図のIDがカレントと一致したら
                                        if(layout.Drawing.ID == drawing.ID)
                                        {
                                            selectTabIndex = i;
                                            break;
                                        }
                                    }
                                }
                            }
                        }
                    }

                    // 部分図が削除された場合
                    // 部分図レイアウト情報表示のクリア
                    int findIndex2 = FindInterlockingDrawingLayoutIdInLayerGroup(deleteingLayoutID);
                    if( findIndex2 >= 0)
                        layerGroupsLayerControls[findIndex2].ClearDrawinglayoutInformation();

                    // タブの切り替え
                    if(selectTabIndex >= 0)
                        layerGroupTabControl.SelectedIndex = selectTabIndex;

                    int selectedTabIndex = layerGroupTabControl.SelectedIndex;
                    Debug.Assert(selectedTabIndex < this.layerGroupsLayerControls.Length);
                    LayerGroupsLayerControl layerGroupsLayerControl = layerGroupsLayerControls[selectedTabIndex];
                    Layer controlLayerGroup = null;

                    // 連動している部分図レイアウトがある場合
                    if(isInterlockingDrawingLayout)
                    {
                        // 部分図レイアウト状態更新
                        controlLayerGroup = layerGroupsLayerControl.GetLayerGroup();
                        layerGroupsLayerControl.UpdateLayoutInfo(controlLayerGroup);
                    }
                    // 連動している部分図レイアウトがなかった場合
                    else
                    {
                        // string msg = "カレントにした部分図レイアウトはレイヤグループと連動してはいません。(このまま作図することは可能です)";
                        // MessageBox.Show(msg, AddInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Information);

                        // 部分図レイアウト状態更新
                        layerGroupsLayerControl.UpdateLayoutInfo(controlLayerGroup);
                    }
                }

                // コントロールの再描画処理
                RefreshControls();
            }

            return;
        }


        /// <summary>
        /// コントロールの再描画処理
        /// </summary>
        private void RefreshControls()
        {
            // カレントレイヤ変更イベント処理中の場合
            if(this.isAppCurrentLayerChanging)
            {
                // タブページのレイヤ状態ボタン再描画
                int selectedTabIndex = layerGroupTabControl.SelectedIndex;
                Debug.Assert(selectedTabIndex < this.layerGroupsLayerControls.Length);
                layerGroupsLayerControls[selectedTabIndex].InvalidateLayerStateButtons();
            }
            // その他 コントロール Refresh フラグ が有効なら
            else if(this.refreshControl)
            {
                Refresh();
            }
        }

        /// <summary>
        /// コンパクト表示モードか
        /// </summary>
        /// <returns></returns>
        public bool IsCompactFormMode()
        {
            return app.addInSettings.compactFormMode;
        }

        /// <summary>
        /// 「その他」メニューが開くときの処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void toolsStripMenuItem_DropDownOpening(object sender, EventArgs e)
        {
            // コンパクト表示モード
            compactModeToolStripMenuItem.Checked = IsCompactFormMode();

            // RootPro CAD のウィンドウの配色を使用するかどうか
            rootProCADVisualStyleToolStripMenuItem.Checked = app.addInSettings.rootProCADVisualStyle;

            // RootPro CAD の表示フォントを使用するかどうか
            rootProCADDisplayFontToolStripMenuItem.Checked = app.addInSettings.rootProCADDisplayFont;

            // ウィンドウの配色と表示フォントの設定は、RootPro CAD 側で Off の場合、
            // アドインでの変更は反映されないので Enabled を false に
            AddInManager addInManager = app.Application.AddInManager;
            rootProCADVisualStyleToolStripMenuItem.Enabled = addInManager.UseVisualStyle;
          　rootProCADDisplayFontToolStripMenuItem.Enabled = addInManager.UseDisplayFont;
        }

        /// <summary>
        /// フォームのコンパクトモード設定
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void compactModeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // コンパクト表示モード
            app.addInSettings.compactFormMode = compactModeToolStripMenuItem.Checked;

            UpdateControlsSize();

            // フローティングバーのサイズを合わせる
            AdjustFloatingBar();
        }

        /// <summary>
        /// RootPro CAD の配色
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void rootProCADVisualStyleToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // メッセージ：次回起動時に反映されます。（キャンセル可能）
            string msg = Properties.Resources.formStyleChangeMsg;
            DialogResult dialogResult = MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OKCancel, MessageBoxIcon.Information);
            if(dialogResult != DialogResult.OK)
                return;

            // RootPro CAD のウィンドウの配色を使用するかどうか
            app.addInSettings.rootProCADVisualStyle = rootProCADVisualStyleToolStripMenuItem.Checked;
        }

        /// <summary>
        /// RootPro CAD のフォント
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void rootProCADDisplayFontToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // メッセージ：次回起動時に反映されます。（キャンセル可能）
            string msg = Properties.Resources.formStyleChangeMsg;
            DialogResult dialogResult = MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OKCancel, MessageBoxIcon.Information);
            if(dialogResult != DialogResult.OK)
                return;

            // RootPro CAD の表示フォントを使用するかどうか
            app.addInSettings.rootProCADDisplayFont = rootProCADDisplayFontToolStripMenuItem.Checked;
        }

        /// <summary>
        /// 指定したレイアウトIDのレイアウトがレイヤグループで関連付けられているか探す
        /// </summary>
        /// <param name="findLayoutID"></param>
        /// <returns>見つかったレイヤグループのリストのインデックス</returns>
        private int FindInterlockingDrawingLayoutIdInLayerGroup(int findLayoutID)
        {
            int findIndex = -1;

            if(findLayoutID < 0)
                return findIndex;
            
            for(int i = 0; i < RcAddInMaxLayerGroupSize; i++)
            {
                Layer layerGroup = layerGroupsLayerControls[i].GetLayerGroup();
                if(layerGroup != null)
                {
                    int layoutID = GetIdFromLayerGroupDescription(layerGroup, ItemName_InterlockingDrawingLayoutID, out int findFirstIndex, out int findTextLength);
                    if(layoutID == findLayoutID)
                    {
                        findIndex = i;
                        break;
                    }
                }
            }

            return findIndex;
        }

        /// <summary>
        /// 指定したレイアIDのレイアがあるタブを探す
        /// </summary>
        /// <param name="findLayerID"></param>
        /// <returns>タブのインデクス</returns>
        private int FindTabIndexIncludeLayerId(int findLayerID)
        {
            if(findLayerID < 0)
                return -1;
            
            for(int i = 0; i < RcAddInMaxLayerGroupSize; i++)
            {
                if(layerGroupsLayerControls[i].IsChildLayerID(findLayerID))
                    return i;
            }

            return -1;

        }

        /// <summary>
        /// レイヤグループのリスト取得
        /// (16個まで)
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="deleteingLayerID"></param>
        /// <returns></returns>
        private List<Layer> GetLayerGroup16List(Document doc, int deleteingLayerID = -1)
        {
            Debug.Assert(doc != null);

            LayerTable layerTable = doc.LayerTable;
            List<Layer> layerGroupList = null;

            // 1階層目のレイヤグループのリスト作成
            int layerGroupCounter = 0;
            LayerCollection layerCollection = layerTable.RootLayer.ChildLayers;
            foreach(Layer layer in layerCollection)
            {
                // グループレイヤの場合
                if(layer.IsGroup)
                {
                    // 削除するレイヤグループはスキップ
                    if(layer.ID == deleteingLayerID)
                        continue;

                    if(layerGroupList == null)
                        layerGroupList = new List<Layer>();

                    layerGroupList.Add(layer);

                    layerGroupCounter++;
                    // レイヤグループ数制限確認
                    if(layerGroupCounter >= RcAddInMaxLayerGroupSize)
                        break;
                }
                else
                {
                    // ルートにレイヤは何もしない
                }
            }

            return layerGroupList;
        }

        /// <summary>
        /// レイヤグループ内のカレントレイヤと連携部分図レイアウトをドキュメントのカレントにする
        /// タブを切り替えるときに行う
        /// </summary>
        /// <param name="selectIndex"></param>
        private void SetCurrentLayerAndLayoutInLayerGroup(int layerGroupIndex)
        {
            Debug.Assert(layerGroupIndex < layerGroupsLayerControls.Length);

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            // Undo, Redo の実行中は何もしない。
            if(doc.UndoManager.IsUndoing || doc.UndoManager.IsRedoing)
                return;

            Layer layerGroup = layerGroupsLayerControls[layerGroupIndex].GetLayerGroup();
            if(layerGroup == null)
                return;

            int findFirstIndex;
            int findTextLength;

            // レイヤグループの説明からカレントレイヤIDを取得してそのカレントレイヤをドキュメントのカレントレイヤにする
            int currentLayerID = GetIdFromLayerGroupDescription(layerGroup, ItemName_CurrentLayerID, out findFirstIndex, out findTextLength);
            // IDが有効かチェック
            if(!AddInTools.IsValidLayerID(doc, currentLayerID))
                currentLayerID = -1;

            // レイヤグループの説明からカレント部分図レイアウトIDを取得してそのカレント部分図レイアウトをドキュメントのカレント部分図レイアウトにする
            int currentLayoutID = GetIdFromLayerGroupDescription(layerGroup, ItemName_InterlockingDrawingLayoutID, out findFirstIndex, out findTextLength);

            if(currentLayerID >= 0 || currentLayoutID >= 0)
            {
                bool beginUndoUnitAtThisFunction = false;
                // UndoUnitの開始
                if(!doc.UndoManager.IsInUndoUnit)
                {
                    doc.UndoManager.BeginUndoUnit();
                    beginUndoUnitAtThisFunction = true;
                }

                if(currentLayerID >= 0)
                {
                    LayerTable layerTable = doc.LayerTable;
                    Layer layer = layerTable.GetLayerByID(currentLayerID);

                    // グループのカレントレイヤをドキュメントにセット
                    if(layer != null)
                    {
                        layerTable.CurrentLayer = layer;
                    }
                }

                if(currentLayoutID >= 0)
                {
                    Layout layout = AddInTools.GetLayoutOfID(doc, currentLayoutID);
                    if(layout != null)
                    {
                        doc.CurrentLayout = layout;
                    }
                }

                // UndoUnit の終了
                if(beginUndoUnitAtThisFunction)
                    doc.UndoManager.EndUndoUnit();
            }

        }

        /// <summary>
        /// 指定したIDの部分図レイアウトが他のレイアグループで連動していないか
        /// </summary>
        /// <param name="layoutID"></param>
        /// <param name="myLayerGroupID"></param>
        /// <returns></returns>
        public bool IsInterlockingDrawinglayoutAtOtherLayerGroup(int layoutID, int myLayerGroupID)
        {
            for(int i = 0; i < RcAddInMaxLayerGroupSize; i++)
            {
                // 指定したレイアグループのはスキップ
                int layerGroupID = layerGroupsLayerControls[i].GetLayerGroupID();
                if(layerGroupID == myLayerGroupID)
                    continue;

                Layer layerGroup = layerGroupsLayerControls[i].GetLayerGroup();
                if (layerGroup == null)
                    continue;

                int currentLayoutID = GetIdFromLayerGroupDescription(layerGroup, ItemName_InterlockingDrawingLayoutID, out int findFirstIndex, out int findTextLength);
                if(currentLayoutID >= 0)
                {
                    if(layoutID == currentLayoutID)
                        return true;
                }
            }

            return false;
        }
         
        /// <summary>
        /// 現在のタブが切り替わる前のイベント処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layerGroupTabControl_Deselecting(object sender, TabControlCancelEventArgs e)
        {
            //
            // メモ：タブの切り替え時のイベントは、_Deselecting -> _Deselected -> _Selecting -> _Selected
            //

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            // タブを切り換えていいか
            // マウスで切り替えようとしていて、そのタブが無効なら切り替えをキャンセル
            int onMouseTabIndex = layerGroupTabControl.onMouseTabIndex;
            if(onMouseTabIndex >= 0)
            {
                Debug.Assert(onMouseTabIndex < this.layerGroupsLayerControls.Length);
                LayerGroupsLayerControl layerGroupsLayerControl = layerGroupsLayerControls[onMouseTabIndex];
                if(!layerGroupsLayerControl.Enabled)
                {
                    e.Cancel = true;
                    return;
                }
            }
            
            // 作図編集ウィンドウで用紙タブが選ばれている場合は切り換えてよい
            // RootPro CAD で部分図タブで直接切り替える場合もよい
            // RootPro CAD でレイヤツリーなどからカレントレイヤを換えた場合もよい
            if(AddInTools.GetActiveTabViewType(app) != AddInTools.RC_TabViewType.RC_TabViewType_Paper && !isAppCurrentDrawingLayoutChanging && !isAppCurrentLayerChanging)
            {
                e.Cancel = true;
                return;
            }

            TabControl tab = (TabControl)sender;

            // 前のタブインデクス保持
            lastSelectTabIndex = tab.SelectedIndex;

            // 切り替わる前のレイヤグループ内のカレントレイヤ情報をレイヤの説明プロパティに保存する
            UpdateLayerGroupDescription(doc, lastSelectTabIndex);
         }

        /// <summary>
        /// 現在のタブが切り替わったときのイベント処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layerGroupTabControl_Deselected(object sender, TabControlEventArgs e)
        {
        }

        /// <summary>
        /// 新しいタブが切り替わろうとしているときのイベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layerGroupTabControl_Selecting(object sender, TabControlCancelEventArgs e)
        {
            TabControl tab = (TabControl)sender;

            // 選択しようとしているタブにレイヤグループが設定されていなければ切り替えキャンセル
            int tabIndex = tab.SelectedIndex;
            LayerGroupsLayerControl layerGroupsLayerControl = layerGroupsLayerControls[tabIndex];
            if(layerGroupsLayerControl.GetLayerGroupID() == -1)
                e.Cancel = true;
        }

        /// <summary>
        /// 新しいタブに切り替わったときのイベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layerGroupTabControl_Selected(object sender, TabControlEventArgs e)
        {
            // タブが切り替わった場合カレントレイヤやレイアウトの切り換えを行う
            int tabIndex = layerGroupTabControl.SelectedIndex;
            if(lastSelectTabIndex != tabIndex)
            {
                // 作図編集ウィンドウの用紙タブが選ばれているときのみ
                if(AddInTools.GetActiveTabViewType(app) == AddInTools.RC_TabViewType.RC_TabViewType_Paper)
                {
                    // カレントになるレイヤと部分図レイアウトのセット
                    SetCurrentLayerAndLayoutInLayerGroup(tabIndex);
                }
                 
                lastSelectTabIndex = tabIndex;
            }
        }

        /// <summary>
        /// 指定したタブのレイヤグループ内のカレントレイヤと連携部分図レイアウトについて説明プロパティを更新する
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="tabIndex"></param>
        private void UpdateLayerGroupDescription(Document doc, int tabIndex)
        {
            Debug.Assert(doc != null);

            // Undo, Redo の実行中は何もしない。
            if(doc.UndoManager.IsUndoing || doc.UndoManager.IsRedoing)
                return;

            int layerGroupsLayerControlsSize = this.layerGroupsLayerControls.Length;
            Debug.Assert(tabIndex < layerGroupsLayerControlsSize);
            LayerGroupsLayerControl layerGroupsLayerControl = layerGroupsLayerControls[tabIndex];
            Layer layerGroup = layerGroupsLayerControl.GetLayerGroup();
            if(layerGroup == null)
                return;

            bool beginUndoUnitAtThisFunction = false;

            // レイヤグループの説明からカレントレイヤID取得
            int currentLayerIdInLayerGroupDescription = GetIdFromLayerGroupDescription(layerGroup, ItemName_CurrentLayerID, out int findFirstIndex, out int findTextLength);
            // IDが有効かチェック
            if(!AddInTools.IsValidLayerID(doc, currentLayerIdInLayerGroupDescription))
                currentLayerIdInLayerGroupDescription = -1;

            // 現在のレイヤグループ内のカレントレイヤについて
            int currentLayerID = layerGroupsLayerControl.GetCheckedCurrentLayerRadioButtonLayerID();
            if(currentLayerIdInLayerGroupDescription != currentLayerID)
            {
                // UndoUnit の開始
                if(!doc.UndoManager.IsInUndoUnit)
                {
                    doc.UndoManager.BeginUndoUnit();
                    beginUndoUnitAtThisFunction = true;
                }

                // レイヤグループの説明にカレントだったレイヤIDを書き込む(カレントがない場合は消す)
                if(doc.UndoManager.IsInUndoUnit)
                    SetIdToLayerGroupDescription(layerGroup, ItemName_CurrentLayerID, currentLayerID);
            }

            // 連携部分図レイアウトについて（RootPro CAD の操作で削除されている場合があるので確認して、なければ消す）(セットは連携設定時にしている)
            int selectLayoutID = GetIdFromLayerGroupDescription(layerGroup, ItemName_InterlockingDrawingLayoutID, out findFirstIndex, out findTextLength);
            if(selectLayoutID >= 0)
            {
                // 部分図レイアウトのIDが現在有効ではなかった場合
                if(!IsAvailabilityDrawingLayoutID(doc, selectLayoutID))
                {
                    // UndoUnit の開始
                    if(!doc.UndoManager.IsInUndoUnit)
                    {
                        doc.UndoManager.BeginUndoUnit();
                        beginUndoUnitAtThisFunction = true;
                    }

                    // レイヤグループの説明から連携部分図レイアウトIDを消す
                    if(doc.UndoManager.IsInUndoUnit)
                        SetIdToLayerGroupDescription(layerGroup, ItemName_InterlockingDrawingLayoutID, -1);
                }
            }

            // UndoUnit の終了
            if(beginUndoUnitAtThisFunction)
                doc.UndoManager.EndUndoUnit();
        }

        /// <summary>
        /// 部分図レイアウトのIDが現在有効か
        /// </summary>
        /// <param name="doc"></param>
        /// <param name=""></param>
        /// <returns></returns>
        private bool IsAvailabilityDrawingLayoutID(Document doc, int selectLayoutID)
        {
            Debug.Assert(doc != null);

            DrawingCollection drawings = doc.Drawings;
            foreach(Drawing drawing in drawings)
            {
                if(drawing is Paper)
                {
                    // 部分図レイアウト
                    LayoutCollection layouts = drawing.Layouts;
                    foreach(Layout layout in layouts)
                    {
                        // 用紙自身のはスキップ
                        if(layout.Drawing is Paper)
                            continue;

                        // 部分図レイアウトのID
                        int layoutDrawingID = layout.ID;
                        if(selectLayoutID == layoutDrawingID)
                            return true;
                    }
                }
            }

            return false;
        }

        /// <summary>
        /// ドキュメントのカレントレイヤのあるレイヤーグループに連携している部分図レイアウトをカレントにする
        /// </summary>
        private void SetCurrentInterlockingDrawinglayoutInLayerGroupIncludeCurrentLayer()
        {
            for(int i = 0; i < RcAddInMaxLayerGroupSize; i++)
            {
                LayerGroupsLayerControl layerGroupsLayerControl = layerGroupsLayerControls[i];
                if(layerGroupsLayerControl.IsDocumentCurrentLayer())
                    SetCurrentLayerAndLayoutInLayerGroup(i);
            }
        }

        /// <summary>
        /// キーボードのキーダウンイベント処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layerGroupTabControl_KeyDown(object sender, KeyEventArgs e)
        {
            // 矢印キーの場合
            if(e.KeyCode == Keys.Up || e.KeyCode == Keys.Down || e.KeyCode == Keys.Right || e.KeyCode == Keys.Left)
            {
                layerGroupTabControl.onMouseTabIndex = -1;// マウスのあるタブインデックスなし
            }
        }

        /// <summary>
        /// キーボードのキーアップイベント処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layerGroupTabControl_KeyUp(object sender, KeyEventArgs e)
        {
            // 矢印キーの場合
            if(e.KeyCode == Keys.Up || e.KeyCode == Keys.Down || e.KeyCode == Keys.Right || e.KeyCode == Keys.Left)
            {
                // 作図編集ウィンドウの用紙タブが選ばれているときのみキーボードでの切り換えができる
                if(AddInTools.GetActiveTabViewType(app) != AddInTools.RC_TabViewType.RC_TabViewType_Paper)
                {
                    // "他のレイヤグループのタブに切り替えたい場合は作図編集ウィンドウで用紙タブを選んでください。";
                    string msg = Properties.Resources.layerGroupTabControl_KeyUp_Msg;
                    MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
        }

        /// <summary>
        /// マウスボタンダウンイベント処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layerGroupTabControl_MouseDown(object sender, MouseEventArgs e)
        {
            // マウスの左ボタンの時
            if(e.Button == MouseButtons.Left)
            {
                // TabControlを取得
                TabControl tab = (TabControl)sender;

                // マウスがどのタブの上か
                int tabIndexUnderMousePoint = layerGroupTabControl.GetTabIndexUnderPoint(e.Location);
                if(tabIndexUnderMousePoint >= 0 && lastSelectTabIndex != tabIndexUnderMousePoint)
                {
                    // 作図編集ウィンドウで用紙タブがアクティブの場合のみマウスでは切り替えられる
                    if(AddInTools.GetActiveTabViewType(app) != AddInTools.RC_TabViewType.RC_TabViewType_Paper)
                    {
                        // "他のレイヤグループのタブに切り替えたい場合は作図編集ウィンドウで用紙タブを選んでください。";
                        string msg = Properties.Resources.layerGroupTabControl_MouseDown;
                        MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Information);
                    }
                }
            }
        }

         
        /// <summary>
        /// レイヤグループのレイヤの状態を変更
        /// </summary>
        /// <param name="childLayers"></param>
        /// <param name="notChangeLayerID">変えないレイヤID　-1:全て変更</param>
        /// <param name="state">表示状態：編集可能、表示のみ、表示しない</param>
        /// <param name="hideFlgValue">-1:表示状態を優先設定をする, 0:隠しレイヤOFF, 1:隠しレイヤ</param>
        public static void SetLayerGroupState(LayerCollection childLayers, int notChangeLayerID = -1, bool notChangeLayerIdIsGroup = false, ShapeState state = ShapeState.Editable, int hideFlgValue = -1)
        {
            foreach(Layer childLayer in childLayers)
            {
                if(childLayer.IsGroup)
                {
                    LayerCollection childLayers2 = childLayer.ChildLayers;
                    SetLayerGroupState(childLayers2, notChangeLayerID, notChangeLayerIdIsGroup, state, hideFlgValue);// 再帰処理
                }
                else
                {
                    // 変更しないレイヤのIDがレイヤグループのものならスキップ
                    if(notChangeLayerIdIsGroup)
                    {
                        if(childLayer.ParentLayer.ID == notChangeLayerID)
                            continue;
                    }

                    if(childLayer.ID != notChangeLayerID)
                    {
                        if(hideFlgValue == -1 && childLayer.State != state)
                            childLayer.State = state;
                        else if(childLayer.Hide == false && hideFlgValue == 1)
                            childLayer.Hide = true;
                        else if(childLayer.Hide == true && hideFlgValue == 0)
                            childLayer.Hide = false;
                    }
                }
            }
        }

        /// <summary>
        /// メニューバーの「図面」が開くとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void drawingStripMenuItem_DropDownOpening(object sender, EventArgs e)
        {
            Document doc = app.ActiveDocument;

            // 「グループの追加」の有効・無効
            bool addLayerGroupToolStripMenuItemEnabled = false;

            // 作図編集ウィンドウの用紙タブがアクティブの場合
            AddInTools.RC_TabViewType tabViewType = AddInTools.GetActiveTabViewType(app);
            if(tabViewType == AddInTools.RC_TabViewType.RC_TabViewType_Paper)
            {
                if(doc != null)
                {
                    // 用紙は1つ
                    if(doc.Drawings.PaperCount == 1)
                    {
                        addLayerGroupToolStripMenuItemEnabled = true;

                        LayerTable layerTable = doc.LayerTable;
                        int layerGroupCount = 0;
                        LayerCollection layerCollection = layerTable.RootLayer.ChildLayers;
                        foreach(Layer layer in layerCollection)
                        {
                            // レイヤグループ
                            if(layer.IsGroup)
                            {
                                layerGroupCount++;
                                if(layerGroupCount >= RcAddInMaxLayerGroupSize)
                                {
                                    addLayerGroupToolStripMenuItemEnabled = false;
                                    break;
                                }
                            }
                            // レイヤ
                            else
                            {
                                // ルート直下にレイアがある場合は無効にする
                                addLayerGroupToolStripMenuItemEnabled = false;
                                break;
                            }
                        }
                    }
                }
            }

            // 「グループの追加」の有効・無効
            addLayerGroupToolStripMenuItem.Enabled = addLayerGroupToolStripMenuItemEnabled;

            // 「図面のチェック」の有効・無効
            checkDocumentToolStripMenuItem.Enabled = (doc != null) ? true:false;
        }

        /// <summary>
        /// メニューバーの「図面」-「新しい図面」を選んだとき
        /// 新規図面を作成追加
        /// このアドインに合わせてルートのレイヤはなく、16個のレイヤを持つレイヤグループを1つ、部分図レイアウトが1つの図面を用意します。
        /// レイヤグループに連携させる部分図レイアウトのスケールは1/1、カレントレイヤはレイヤーグループ内の最初のレイヤになります。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void newDocumentToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // 新規図面を作成追加
            Document doc = app.Documents.Add("");
            if(doc == null)
                return;

            Document activeDoc = app.ActiveDocument;
            Debug.Assert(activeDoc.Equals(doc));

            // コントロール Refresh 有効フラグ off
            bool oldRefreshControl = this.refreshControl;
            this.refreshControl = false;

            // 一旦部分図を削除する(UndoUnitはリセットされるのでUndoUnitの開始はしなくてよい)
            DrawingCollection drawings = doc.Drawings;
            int drawingsCount = drawings.Count;
            for(int i = drawingsCount - 1; i >= 0; i--)
            {
                Drawing drawing = drawings[i];
                if(drawing is Part)
                    drawings.RemoveAt(i);
            }

            // レイヤグループと部分図の追加
            string groupName = "";
            AddLayerGroupAndDrawingLayout(groupName);

            // ルート直下のレイヤを削除
            LayerTable layerTable = doc.LayerTable;
            int rootLayerCount = 0;
            LayerCollection layers = layerTable.RootLayer.ChildLayers;
            foreach(Layer layer in layers)
            {
                if(!layer.IsGroup)
                    rootLayerCount++;
            }

            if(rootLayerCount > 0)
            {
                // UndoUnitの開始
                doc.UndoManager.BeginUndoUnit();

                foreach(Layer layer in layers)
                {
                    // 削除
                    if(!layer.IsGroup)
                        layers.RemoveByID(layer.ID);
                }

                // UndoUnitの終了
                doc.UndoManager.EndUndoUnit();
            }

            // すべての部分図タブを閉じる
            CloseAllPartDrawingView();

            // 新規図面を開いた状態とてここまでのUndoUnit履歴はクリア
            doc.UndoManager.Clear();

            // コントロール Refresh 有効フラグを戻す
            this.refreshControl = oldRefreshControl;

            UpdateFormControlState();
        }

        /// <summary>
        /// メニューバーの「図面」-「レイヤグループと部分図の追加」を選んだとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void addLayerGroupToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // レイヤグループと部分図の追加
            string groupName = "";
            AddLayerGroupAndDrawingLayout(groupName);
        }

        /// <summary>
        /// レイヤグループと部分図の追加
        /// </summary>
        /// <param name="groupName"></param>
        private void AddLayerGroupAndDrawingLayout(string groupName)
        {
            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            // 作図編集ウィンドウの用紙タブがアクティブの場合のみ実行可能
            AddInTools.RC_TabViewType tabViewType = AddInTools.GetActiveTabViewType(app);
            if(tabViewType != AddInTools.RC_TabViewType.RC_TabViewType_Paper)
                return;

            LayerTable layerTable = doc.LayerTable;
            int layerGroupCount = 0;
            LayerCollection layerCollection = layerTable.RootLayer.ChildLayers;
            foreach(Layer layer in layerCollection)
            {
                if(layer.IsGroup)
                    layerGroupCount++;
            }

            if(layerGroupCount >= RcAddInMaxLayerGroupSize)
                return;

            // UndoUnitの開始
            doc.UndoManager.BeginUndoUnit();

            // コントロール Refresh 有効フラグ off
            bool oldRefreshControl = this.refreshControl;
            this.refreshControl = false;

            // 新しい名前レイヤグループの追加
            if(groupName == "")
                groupName = AddInTools.GetNewLayerGroupName(layerTable, RcAddInMaxLayerGroupSize);

            Layer newLayerGroup = layerCollection.AddLayerGroup(groupName);
            // レイヤも16個追加して、最初のをカレントにする
            LayerCollection childLayers = newLayerGroup.ChildLayers;
            Layer childLayer0 = childLayers.AddLayer("0");
            layerTable.CurrentLayer = childLayer0;// 最初のレイヤをカレントに
            for (int i = 1; i < RcAddInChildLayerSize; i++)
            {
                string newLayerName = i.ToString("X"); // 16進数文字で
                childLayers.AddLayer(newLayerName);
            }

            // コントロール Refresh 有効フラグを戻す
            this.refreshControl = oldRefreshControl;

            // レイヤグループの説明にカレントのレイヤIDを書き込む
            SetIdToLayerGroupDescription(newLayerGroup, ItemName_CurrentLayerID, childLayer0.ID);

            // 部分図レイアウトの追加
            string drawingName = AddInTools.GetNewDrawingName(doc.Drawings, RcAddInMaxLayerGroupSize);
            Layout layout = AddDrawingLayout(drawingName);
            if(layout != null)
            {
                // 追加した部分図は表示のみにする
                Drawing drawing = layout.Drawing;
                drawing.State = ShapeState.Visible;

                int newLayoutID = layout.ID;

                // 追加した部分図の部分図レイアウトをスケール連携
                SetIdToLayerGroupDescription(newLayerGroup, ItemName_InterlockingDrawingLayoutID, newLayoutID);

                // 追加した部分図レイアウトをカレントに
                doc.CurrentLayout = layout;
            }

            // UndoUnitの終了
            doc.UndoManager.EndUndoUnit();
        }

        /// <summary>
        /// 現在の用紙に新しい部分図とその部分図の部分図レイアウトを追加
        /// </summary>
        /// <param name="drawingName"></param>
        /// <returns>int 新しい部分図レイアウトID</returns>
        private Layout AddDrawingLayout(string drawingName)
        {
            Layout layout = null;

            Document doc = app.ActiveDocument;
            if(doc == null)
                return null;

            TabView tabView = app.ActiveTabView;
            if(tabView == null)
                return null;

            // 部分図ビューの場合
            if(tabView is DrawingView)
            {
                DrawingView drawingView = (DrawingView)tabView;

                // 現在のタブが用紙の場合
                if(drawingView.Drawing is Paper)
                {
                    // 用紙
                    Drawing paperDrawing = drawingView.Drawing;

                    // 部分図コレクション
                    DrawingCollection drawings = doc.Drawings;

                    // UndoUnitの開始
                    bool beginUndoUnitAtThisFunction = false;
                    if(!doc.UndoManager.IsInUndoUnit)
                    {
                        doc.UndoManager.BeginUndoUnit();
                        beginUndoUnitAtThisFunction = true;
                    }

                    // 部分図の追加
                    Drawing newPartDrawing = drawings.AddPart(drawingName);

                    // 部分図名
                    drawingName = newPartDrawing.Name;

                    // 用紙に部分図レイアウトを追加する
                    LayoutCollection layouts = paperDrawing.Layouts;
                    string layoutName = "Layout (" + drawingName + ")";
                    layout = layouts.Add(layoutName, newPartDrawing);

                    // 用紙の上になるように表示順を変更
                    int paperDrawingLayoutIndex = AddInTools.GetPaperDrawingLayoutIndex(layouts);
                    if(paperDrawingLayoutIndex >= 0)
                        layout.Index = paperDrawingLayoutIndex;

                    // 追加した部分図のタブビューを閉じる
                    ClosePartDrawingView(newPartDrawing.ID);

                    // UndoUnitの終了
                    if(beginUndoUnitAtThisFunction)
                        doc.UndoManager.EndUndoUnit();
                }
            }

            return layout;
        }

        /// <summary>
        /// 「図面のチェック」が選ばれたとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void checkDocumentToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            // "このアドインを使って作図する場合の図面は、次の条件が満たされていることが推奨されます。、、、
            string msg = String.Format(Properties.Resources.checkDocumentMsg, RcAddInMaxLayerGroupSize, RcAddInMaxChildLayerSize);
            DialogResult result = MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Information);
            if(result != DialogResult.Yes)
                return;

            // ドキュメントチェック
            if(!IsConvertableDocument(doc, out string strErr))
                msg = strErr;
            else
                msg = Properties.Resources.checkDocumentNoProblem;// "特に問題はありませんでした。";

            MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        /// <summary>
        /// コンテキストメニューが開くとき選んだタブのインデックス更新
        /// </summary>
        /// <param name="menu"></param>
        /// <returns></returns>
        private bool UpdateContextMenuSelectTabIndex(ContextMenuStrip menu)
        {
            // マウスカーソルの位置(画面座標)
            Point mousePosition = Control.MousePosition;
            // クライアント座標に変換
            Point clientPoint = menu.SourceControl.PointToClient(mousePosition);

            // マウスカーソルの位置のタブのインデックスを取得して更新保持
            this.contextMenuSelectTabIndex = layerGroupTabControl.GetTabIndexUnderPoint(clientPoint);
            if(this.contextMenuSelectTabIndex != -1)
                return true;

            return false;
        }

        /// <summary>
        /// コンテキストメニューオープン時に選んだタブのレイヤグループID取得
        /// </summary>
        /// <returns>レイヤーグループID</returns>
        private int GetContextMenuSelectLayerGroupID()
        {
            if(this.contextMenuSelectTabIndex >= 0)
            {
                // コンテキストメニューオープン時のコントロールのインデックスからどのレイヤグループかを決める
                int changeLayerGroupID = layerGroupsLayerControls[this.contextMenuSelectTabIndex].GetLayerGroupID();
                if(changeLayerGroupID != -1)
                    return changeLayerGroupID;
            }

            return -1;
        }


        /// <summary>
        /// コンテキストメニューオープン時に選んだタブのレイヤグループに連携しいる部分図レイアウト取得
        /// </summary>
        /// <returns></returns>
        private Layout GetContextMenuSelectLayerGroupInterlockingDrawingLayout()
        {
            Layout interlockingDrawingLayout = null;
            if(this.contextMenuSelectTabIndex >= 0)
            {
                interlockingDrawingLayout = layerGroupsLayerControls[this.contextMenuSelectTabIndex].GetInterlockingDrawingLayout();
            }

            return interlockingDrawingLayout;
        }

        /// <summary>
        /// コンテキストメニューが開くときのイベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void groupContextMenuStrip_Opening(object sender, System.ComponentModel.CancelEventArgs e)
        {
            // コンテキストメニューが開くとき選んだタブのインデックス更新
            UpdateContextMenuSelectTabIndex((ContextMenuStrip)sender);

            // コンテキストメニューが開くとき選んだタブにあるレイヤグループID
            int contextMenuSelectLayerGroupID = GetContextMenuSelectLayerGroupID();
            // 作図編集ウィンドウで用紙タブが選ばれていない場合もキャンセル
            if(contextMenuSelectLayerGroupID < 0 || AddInTools.GetActiveTabViewType(app) != AddInTools.RC_TabViewType.RC_TabViewType_Paper)
            {
                e.Cancel = true;
                return;
            }

            Document doc = app.ActiveDocument;
            if(doc == null)
            {
                e.Cancel = true;
                return;
            }

            // 現在選ばれているタブのインデックス
            int selectedTabIndex = layerGroupTabControl.SelectedIndex;
           
            LayerTable layerTable = doc.LayerTable;
            Layer layerGroup = layerTable.GetLayerByID(contextMenuSelectLayerGroupID);
            Debug.Assert(layerGroup.IsGroup);

            // 「レイヤグループ」-「レイヤの追加」メニュー
            // 現在のレイヤグループでのみ選択可能
            bool enabledFlg = false;
            if(selectedTabIndex == this.contextMenuSelectTabIndex)
            {
                // 子レイヤが最大数を超えていない場合
                LayerCollection layerCollection = layerGroup.ChildLayers;
                if(layerCollection.Count < RcAddInMaxChildLayerSize)
                    enabledFlg = true;
            }
            addLayerToolStripMenuItem.Enabled = enabledFlg;

            // 「グループの削除」メニュー
            // ルートにあるレイヤグループが2つ以上の場合有効
            // 且つ、削除しようとしているレイヤグループ以外のレイヤグループにレイヤ1つ以上あること
            Layer otherLayer = AddInTools.FindOtherLayerGroupsLayer(layerTable, layerGroup);
            if(AddInTools.GetRootLayerGroupCount(layerTable) >= 2 && otherLayer != null)
                deleteToolStripMenuItem.Enabled = true;
            else
                deleteToolStripMenuItem.Enabled = false;


            // 「部分図」に関する状態
            Layout interlockingDrawingLayout = GetContextMenuSelectLayerGroupInterlockingDrawingLayout();
            if(interlockingDrawingLayout != null)
            {
                drawingToolStripMenuItem.Enabled = true;
                drawinglayoutSettingToolStripMenuItem.Enabled = true;

                Drawing drawing = interlockingDrawingLayout.Drawing;
                ShapeState shapeState = drawing.State;

                // 編集可能
                editLayoutToolStripMenuItem.Checked = shapeState == ShapeState.Editable;

                // 表示のみ
                lockLayoutToolStripMenuItem.Checked = shapeState == ShapeState.Visible;

                // 表示しない
                undisplayLayoutToolStripMenuItem.Checked = shapeState == ShapeState.NotVisible;

                // 印刷する
                printDrawingToolStripMenuItem.Checked = drawing.Printable;
            }
            else
            {
                drawingToolStripMenuItem.Enabled = false;
                drawinglayoutSettingToolStripMenuItem.Enabled = false;
            }
        }

        /// <summary>
        /// コンテキストメニューが閉じたときのイベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void groupContextMenuStrip_Closed(object sender, ToolStripDropDownClosedEventArgs e)
        {
            if(layerGroupTabControl.onMouseTabIndex >= 0)
            {
                // タブ部分の再描画
                Rectangle tabItemRect = layerGroupTabControl.GetTabRect(layerGroupTabControl.onMouseTabIndex);
                layerGroupTabControl.Invalidate(tabItemRect);
            }

            // マウスがタブコントロールの領域内に入ったときのタブのインデックス初期化
            layerGroupTabControl.onMouseTabIndex = -1;
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「編集可能」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void editLayerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LayerToolStripMenuItemClick(ShapeState.Editable);
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「表示のみ」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lockLayerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LayerToolStripMenuItemClick(ShapeState.Visible);
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「表示しない」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void undisplayLayerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LayerToolStripMenuItemClick(ShapeState.NotVisible);
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「隠しレイヤ」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void hideLayerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LayerToolStripMenuItemClick(0, 1);
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「隠しレイヤ OFF」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void hideLayerOffToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LayerToolStripMenuItemClick(0, 0);
        }

        /// <summary>
        /// コンテキストメニューからの状態変更共通処理
        /// </summary>
        /// <param name="state"></param>
        /// <param name="hideFlgValue"></param>
        private void LayerToolStripMenuItemClick(ShapeState state = ShapeState.Editable, int hideFlgValue = -1)
        {
            // このレイヤグループのID
            int changeLayerGroupID = GetContextMenuSelectLayerGroupID();
            if(changeLayerGroupID < 0)
                return;

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            LayerTable layerTable = doc.LayerTable;
            Layer layerGroup = layerTable.GetLayerByID(changeLayerGroupID);
            Debug.Assert(layerGroup.IsGroup);

            LayerCollection layerCollection = layerGroup.ChildLayers;
            if(layerCollection.Count == 0)
                return;

            // UndoUnitの開始
            doc.UndoManager.BeginUndoUnit();

            // レイヤグループのレイヤの状態を変更
            SetLayerGroupState(layerCollection, -1, false, state, hideFlgValue);

            // UndoUnitの終了
            doc.UndoManager.EndUndoUnit();
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「他のレイヤ」について変更が選ばれた時の処理
        /// </summary>
        /// <param name="state">表示状態：編集可能、表示のみ、表示しない</param>
        /// <param name="hideFlgValue">-1:表示状態を優先設定をする, 0:隠しレイヤOFF, 1:隠しレイヤ</param>
        private void OtherLayerToolStripMenuItemClick(ShapeState state = ShapeState.Editable, int hideFlgValue = -1)
        {
            // このレイヤグループのID
            int notChangeLayerGroupID = GetContextMenuSelectLayerGroupID();
            if(notChangeLayerGroupID < 0)
                return;

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;            

            // 子レイヤのレイヤコレクション取得
            LayerTable layerTable = doc.LayerTable;
            LayerCollection layerCollection = layerTable.RootLayer.ChildLayers;

            // UndoUnitの開始
            doc.UndoManager.BeginUndoUnit();

            // 他のレイヤ状態を変更
            SetLayerGroupState(layerCollection, notChangeLayerGroupID, true, state, hideFlgValue);

            // UndoUnitの終了
            doc.UndoManager.EndUndoUnit();
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「他のレイヤを」-「編集可能」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void otherEditLayerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OtherLayerToolStripMenuItemClick(ShapeState.Editable);
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「他のレイヤを」-「表示のみ」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void otherLockLayerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OtherLayerToolStripMenuItemClick(ShapeState.Visible);
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「他のレイヤを」-「表示しない」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void otherUndisplayLayerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OtherLayerToolStripMenuItemClick(ShapeState.NotVisible);
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「他のレイヤを」-「隠しレイヤ」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void otherHideLayerToolStripMenuItem_Click(object sender, EventArgs e)
        {
           OtherLayerToolStripMenuItemClick(0, 1);
        }

        /// <summary>
        /// コンテキストメニューから「レイヤグループ」-「他のレイヤを」-「隠しレイヤ OFF」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void otherHideLayerOffToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OtherLayerToolStripMenuItemClick(0, 0);
        }

        /// <summary>
        /// コンテキストメニューからレイヤの「レイヤグループ」-「レイヤの追加」を選んだとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void addLayerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // このレイヤグループのID
            int changeLayerGroupID = GetContextMenuSelectLayerGroupID();
            if(changeLayerGroupID < 0)
                return;

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            LayerTable layerTable = doc.LayerTable;
            Layer layerGroup = layerTable.GetLayerByID(changeLayerGroupID);
            Debug.Assert(layerGroup.IsGroup);

            LayerCollection layerCollection = layerGroup.ChildLayers;
            if(layerCollection.Count >= RcAddInMaxChildLayerSize)
                return;

             // UndoUnitの開始
            doc.UndoManager.BeginUndoUnit();

            // 新しい名前でレイヤの追加
            Layer newLayer = layerCollection.AddLayer("");
            // カレントレイヤにする
            layerTable.CurrentLayer = newLayer;

            // UndoUnitの終了
            doc.UndoManager.EndUndoUnit();

        }

        /// <summary>
        /// コンテキストメニューから「部分図」-「編集可能」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void editDrawingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DrawingToolStripMenuItemClick(ShapeState.Editable);
        }

        /// <summary>
        /// コンテキストメニューから「部分図」-「表示のみ」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lockDrawingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DrawingToolStripMenuItemClick(ShapeState.Visible);
        }

        /// <summary>
        /// コンテキストメニューから「部分図」-「表示しない」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void undisplayDrawingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DrawingToolStripMenuItemClick(ShapeState.NotVisible);
        }

        private void DrawingToolStripMenuItemClick(ShapeState state = ShapeState.Editable)
        {
            // このレイヤグループと連携する部分図レイアウト
            Layout interlockingDrawingLayout = GetContextMenuSelectLayerGroupInterlockingDrawingLayout();
            if(interlockingDrawingLayout == null)
                return;

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            // UndoUnitの開始
            doc.UndoManager.BeginUndoUnit();

            Drawing drawing = interlockingDrawingLayout.Drawing;
            drawing.State = state;

            // UndoUnitの終了
            doc.UndoManager.EndUndoUnit();
        }

        /// <summary>
        /// コンテキストメニューから「スケールの設定...」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void drawinglayoutSettingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // コンテキストメニューオープン時のコントロールのインデックスからどのレイヤグループかを決める
            if(this.contextMenuSelectTabIndex >= 0)
            {
                // 部分図レイアウトのスケール連携設定
                layerGroupsLayerControls[this.contextMenuSelectTabIndex].InterlockingDrawinglayoutSetting();
            }
        }

        /// <summary>
        /// コンテキストメニューから「部分図」-「他の部分図を」-「編集可能」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void otherEditDrawingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetOtherDrawingState(ShapeState.Editable);
        }

        /// <summary>
        /// コンテキストメニューから「部分図」-「他の部分図を」-「表示のみ」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void otherLockDrawingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetOtherDrawingState(ShapeState.Visible);
        }

        /// <summary>
        /// コンテキストメニューから「部分図」-「他の部分図を」-「表示しない」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void otherUndisplayDrawingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetOtherDrawingState(ShapeState.NotVisible);
        }

        /// <summary>
        /// 他の部分図の状態セット
        /// </summary>
        /// <param name="state"></param>
        private void SetOtherDrawingState(ShapeState state)
        {
            // このレイヤグループと連携する部分図レイアウト
            Layout notChangeInterlockingDrawingLayout = GetContextMenuSelectLayerGroupInterlockingDrawingLayout();
            if(notChangeInterlockingDrawingLayout == null)
                return;

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;            

            // 子レイヤのレイヤコレクション取得
            LayerTable layerTable = doc.LayerTable;
            LayerCollection layerCollection = layerTable.RootLayer.ChildLayers;

            // UndoUnitの開始
            doc.UndoManager.BeginUndoUnit();

            // 他のレイヤ状態を変更
            int layerGroupsLayerControlsSize = layerGroupsLayerControls.Length;
            for(int i = 0; i < layerGroupsLayerControlsSize; i++)
            {
                Layout interlockingDrawingLayout = layerGroupsLayerControls[i].GetInterlockingDrawingLayout();
                if(interlockingDrawingLayout != null)
                {
                    if(interlockingDrawingLayout.Equals(notChangeInterlockingDrawingLayout))
                        continue;

                    // 状態の変更
                    Drawing drawing = interlockingDrawingLayout.Drawing;
                    drawing.State = state;
                }
            }

            // UndoUnitの終了
            doc.UndoManager.EndUndoUnit();
        }

        /// <summary>
        /// コンテキストメニューから部分図の「印刷する」が選ばれたとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void printDrawingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            bool printable = printDrawingToolStripMenuItem.Checked;

            // このレイヤグループと連携する部分図レイアウト
            Layout interlockingDrawingLayout = GetContextMenuSelectLayerGroupInterlockingDrawingLayout();
            if(interlockingDrawingLayout == null)
                return;

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            // UndoUnitの開始
            doc.UndoManager.BeginUndoUnit();

            Drawing drawing = interlockingDrawingLayout.Drawing;
            drawing.Printable = printable;

            // UndoUnitの終了
            doc.UndoManager.EndUndoUnit();
        }

        /// <summary>
        /// コンテキストメニューから「名前の変更...」が選ばれた場合
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void editGroupNameToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // このレイヤグループのID
            int changeLayerGroupID = GetContextMenuSelectLayerGroupID();
            if(changeLayerGroupID < 0)
                return;

            // このレイヤグループと連携する部分図レイアウト
            Layout interlockingDrawingLayout = GetContextMenuSelectLayerGroupInterlockingDrawingLayout();

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            LayerTable layerTable = doc.LayerTable;
            Layer layer = layerTable.GetLayerByID(changeLayerGroupID);
            Debug.Assert(layer.IsGroup);

            Drawing drawing = null;
            if(interlockingDrawingLayout != null)
                drawing = interlockingDrawingLayout.Drawing;
 
            // 名前変更のダイアログボックス表示
            EditGroupNameForm dlg = new EditGroupNameForm(doc, layer, interlockingDrawingLayout);
            if(dlg.ShowDialog() == DialogResult.OK)
            {
                // レイヤーグループ名、部分図名、部分図レイアウト名（事前に同じ名前がないかチェックすること）
                string layerGroupName = dlg.GetLayerGroupName();
                string oldDrawingName = "";
                string drawingName = "";
                string drawingLayoutName = "";
                if(interlockingDrawingLayout != null)
                {
                    oldDrawingName = drawing.Name;
                    drawingName = dlg.GetDrawingName();

                    drawingLayoutName = dlg.GetDrawingLayoutName();
                }

                if(layer.Name != layerGroupName || oldDrawingName != drawingName)
                {
                    // UndoUnitの開始
                    doc.UndoManager.BeginUndoUnit();

                    // 名称変更
                    // レイヤーグループ名
                    if(layer.Name != layerGroupName)
                        layer.Name = layerGroupName;

                    if(interlockingDrawingLayout != null)
                    {
                        // 部分図名
                        if(drawing.Name != drawingName)
                            drawing.Name = drawingName;

                        // 部分図レイアウト名
                        if(interlockingDrawingLayout.Name != drawingLayoutName)
                            interlockingDrawingLayout.Name = drawingLayoutName;
                    }

                    // UndoUnitの終了
                    doc.UndoManager.EndUndoUnit();
                }
            }
        }

        /// <summary>
        /// コンテキストメニューからレイヤの「レイヤグループと部分図の削除」を選んだとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void deleteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // このレイヤグループのID
            int changeLayerGroupID = GetContextMenuSelectLayerGroupID();
            if(changeLayerGroupID < 0)
                return;

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            LayerTable layerTable = doc.LayerTable;
            Layer layerGroup = layerTable.GetLayerByID(changeLayerGroupID);
            Debug.Assert(layerGroup.IsGroup);

            // ルートにあるレイヤグループが2つ以上の場合有効
            // 且つ、削除しようとしているレイヤグループ以外のレイヤグループにレイヤ1つ以上あること
            Layer otherLayerGroupsLayer = AddInTools.FindOtherLayerGroupsLayer(layerTable, layerGroup);
            if(!(AddInTools.GetRootLayerGroupCount(layerTable) >= 2 && otherLayerGroupsLayer != null))
                return;

            // 削除していいかチェック
            // 作図部品定義内のフィーチャで使用されているレイヤは削除不可
            // 作図グループ内のフィーチャで使用されているレイヤは削除不可
            if(!AddInTools.IsDeleteableLayer(doc, layerGroup, false))
                return;

            // 図形に使われているレイヤだった場合削除していいか問い合わせ
            bool isUseLayerShape = AddInTools.IsUseLayerShape(doc, layerGroup);
            // 連携する部分図レイアウト
            // レイヤグループの説明文字列から指定した項目名のIDを取得
            int currentLayoutID = GetIdFromLayerGroupDescription(layerGroup, ItemName_InterlockingDrawingLayoutID, out int dummyIndex1, out int dummyIndex2);
            Layout layout = AddInTools.GetLayoutOfID(doc, currentLayoutID);
            if(isUseLayerShape || layout != null)
            {
                string msg = Properties.Resources.doYouWantToDeleteThisGroup;
                msg += "\n\n";
                msg += Properties.Resources.layerGroup + ":" + layerGroup.Name;// レイヤグループ：ABC
                msg += "\n";
                msg += Properties.Resources.partDrawing + ":" + layout.Drawing.Name;// 部分図:ABC
                msg += "\n\n";
                msg += Properties.Resources.deleteToolStripMenuItem_Click_Msg_1;// "レイヤグループと部分図は削除されます。グループ内のレイヤを使っている図形も削除されます。";
                msg += "\n";
                msg += Properties.Resources.deleteToolStripMenuItem_Click_Msg_2;// "Undo Redo 履歴も削除されので元に戻すことはできなくなります。注意してください。";
                DialogResult dialogResult = MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2);
                if(dialogResult != DialogResult.Yes)
                    return;
            }

            // コントロール Refresh 有効フラグ off
            bool oldRefreshControl = this.refreshControl;
            this.refreshControl = false;

            // UndoUnitの開始
            doc.UndoManager.BeginUndoUnit();
            bool beginUndoUnitAtThisFunction = true;

            // カレントレイヤが含まれている場合
            Layer otherLayerGroup = null;
            if(AddInTools.IsCurrentLayer(doc, layerGroup))
            {
                // 他のレイヤが含まれているレイヤーグループ
                otherLayerGroup = otherLayerGroupsLayer.ParentLayer;
                // グループ内でカレントにするレイヤのID
                int currentLayerID = GetIdFromLayerGroupDescription(otherLayerGroup, ItemName_CurrentLayerID, out int damyIndex1, out int damyIndex2);
                // IDが有効かチェック
                if(!AddInTools.IsValidLayerID(doc, currentLayerID))
                    currentLayerID = -1;

                // 他のグループのカレントレイヤを次のカレントレイヤに
                if(currentLayerID >= 0)
                {
                    Layer layer2 = layerTable.GetLayerByID(currentLayerID);
                    if(layer2 != null)
                        otherLayerGroupsLayer = layer2;
                }
                               
                // カレントレイヤ変更（カレントレイヤは削除できないため）
                layerTable.CurrentLayer = otherLayerGroupsLayer;
            }

            // ルートの子レイヤグループ
            LayerCollection childLayers = layerTable.RootLayer.ChildLayers;
            // 削除
            childLayers.RemoveByID(layerGroup.ID);

            // 部分図削除
            if(layout != null)
            {
                // 親の部分図を取得
                // Drawing parentDrawing = layout.ParentDrawing;
                // 削除実行
                // parentDrawing.Layouts.Remove(layout);// 部分図レイアウトの削除

                // 部分図を削除する（注意：部分図を削除の削除をすると RootPro CAD の Undo 履歴はクリアされる仕様）
                if(beginUndoUnitAtThisFunction) // UndoUnitの終了
                {
                    doc.UndoManager.EndUndoUnit();
                    beginUndoUnitAtThisFunction = false;
                }

                Drawing drawing = layout.Drawing;
                DrawingCollection drawings = doc.Drawings;
                drawings.Remove(drawing);
            }

            // UndoUnitの終了
            if(beginUndoUnitAtThisFunction)
                doc.UndoManager.EndUndoUnit();

            // コントロール Refresh 有効フラグを戻す
            this.refreshControl = oldRefreshControl;

            // コントロールの再描画処理(タブの再描画をさせるため)
            RefreshControls();
        }

        /// <summary>
        /// 「スケール連携設定」
        /// </summary>
        private void interlockingDrawinglayoutSetting()
        {
            // 今のタブのレイヤグループ
            int tabIndex = layerGroupTabControl.SelectedIndex;
            if (tabIndex < 0)
                return;

            int layerGroupsLayerControlsSize = this.layerGroupsLayerControls.Length;
            Debug.Assert(tabIndex < layerGroupsLayerControlsSize);
            LayerGroupsLayerControl layerGroupsLayerControl = layerGroupsLayerControls[tabIndex];

            // 部分図レイアウト連携設定
            layerGroupsLayerControl.InterlockingDrawinglayoutSetting();

            // 状態再描画
            Layer layerGroup = layerGroupsLayerControl.GetLayerGroup();
            layerGroupsLayerControl.UpdateLayerState(layerGroup);
        }

        /// <summary>
        /// タブコントロール上でのマウス移動イベント処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layerGroupTabControl_MouseMove(object sender, MouseEventArgs e)
        {
            // ツールチップのアクティブ設定（タブ上の移動に合わせる）
            Point cp = new Point(e.X, e.Y);
            int tabIndex = layerGroupTabControl.GetTabIndexUnderPoint(cp);
            if(groupToolTip.Active && tabIndex >= 0 && lastToolTipSelectTabIndex >= 0 && tabIndex != lastToolTipSelectTabIndex)
            {
                groupToolTip.Active = false;
                lastToolTipSelectTabIndex = -1;
            }
            else
            {
                groupToolTip.Active = true;
            }

            // マウスが前と違うタブに入ったとき
            // タブの部分を再描画させる
            if(layerGroupTabControl.onMouseTabIndex != tabIndex)
            {
                Debug.Assert(sender is TabControl);
                TabControl tab = (TabControl)sender;

                if(layerGroupTabControl.onMouseTabIndex >= 0)
                {
                    Rectangle tabItemRect = tab.GetTabRect(layerGroupTabControl.onMouseTabIndex);
                    tab.Invalidate(tabItemRect);
                }

                if(tabIndex >= 0)
                {
                    Rectangle tabItemRect = tab.GetTabRect(tabIndex);
                    tab.Invalidate(tabItemRect);
                }

            }

            // マウスがタブコントロールの領域内に入ったときのタブのインデックス更新
            layerGroupTabControl.onMouseTabIndex = tabIndex;
        }

        /// <summary>
        /// タブコントロールの領域にマウスポインターが出たときのイベント処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layerGroupTabControl_MouseLeave(object sender, EventArgs e)
        {
            Debug.Assert(sender is TabControl);
            TabControl tab = (TabControl)sender;
            if(layerGroupTabControl.onMouseTabIndex >= 0)
            {
                // タブ部分の再描画
                Rectangle tabItemRect = tab.GetTabRect(layerGroupTabControl.onMouseTabIndex);
                tab.Invalidate(tabItemRect);
            }

            // マウスがタブコントロールの領域内に入ったときのタブのインデックス初期化
            layerGroupTabControl.onMouseTabIndex = -1;
        }


        /// <summary>
        /// ツールチップ表示のときの処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void groupToolTip_Popup(object sender, PopupEventArgs e)
        {
            Control control = e.AssociatedControl;
            if(control == null)
                return;

            if(control.Equals(layerGroupTabControl))
            {
                TabControl tab = (TabControl)(control);

                Point sp = Cursor.Position;
                Point cp = tab.PointToClient(sp);

                int tabIndex = layerGroupTabControl.GetTabIndexUnderPoint(cp);
                if(tabIndex >= 0)
                {
                    if(tabIndex >= 0 && tabIndex < layerGroupsLayerControls.Length)
                    {
                        // レイヤグループ名取得
                        string layerGroupName = layerGroupsLayerControls[tabIndex].GetLayerGroupName();
                        string toolTipText = "";
                        if(string.IsNullOrEmpty(layerGroupName))
                        {
                            toolTipText = Properties.Resources.layerGroup + ":" +　Properties.Resources.none;// "レイヤグループ:なし";
                        }
                        else
                        {
                            toolTipText = Properties.Resources.layerGroupName + ":" + layerGroupName;// "レイヤグループ名:" + layerGroupName;
                        }

                        // 部分図情報
                        Layout layout = layerGroupsLayerControls[tabIndex].GetInterlockingDrawingLayout();
                        if(layout != null)
                        {
                            string drawingName = layout.Drawing.Name;
                            string layoutName = layout.Name;
                            toolTipText += "\n";
                            toolTipText += Properties.Resources.partDrawingName + ":" + drawingName;// "部分図名:" + drawingName;
                            toolTipText += "\n";
                            toolTipText += Properties.Resources.partDrawingLayoutName + ":" + layoutName;// "部分図レイアウト名:" + layoutName;

                            // スケール
                            double scaleXNumerator = layout.ScaleXNumerator;      // 分子値
                            double scaleXDenominator = layout.ScaleXDenominator;  // 分母値
                            // スケール
                            string scaleText = Properties.Resources.scale + ": " + scaleXNumerator.ToString() + " / " + scaleXDenominator.ToString();
                            toolTipText += "\n" + scaleText;
                        }

                        // ツールチップのサイズ
                        Size orgToolTipSize = TextRenderer.MeasureText(toolTipText, this.Font);
                        int toolTipMargin = this.Font.Height;
                        Size toolTipSize = new Size(orgToolTipSize.Width + toolTipMargin , orgToolTipSize.Height + toolTipMargin);
                        e.ToolTipSize = toolTipSize;

                        // ツールチップの文字更新
                        if(groupToolTip.GetToolTip(e.AssociatedControl) != toolTipText)
                            groupToolTip.SetToolTip(e.AssociatedControl, toolTipText);

                        // ツールチップが表示されたときのタブのインデックス更新
                        lastToolTipSelectTabIndex = tabIndex;
                    }
                }
            }
        }

        /// <summary>
        /// ツールチップのオーナードロー
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void groupToolTip_Draw(object sender, DrawToolTipEventArgs e)
        {
            e.DrawBackground();
            e.DrawBorder();

            int toolTipMargin = this.Font.Height;
            int margin = (int)(toolTipMargin * 0.5);
            Rectangle rText = e.Bounds;
            rText.Offset(margin, margin);

            Color textColor = groupToolTip.ForeColor;
            TextFormatFlags textFormatFlags = TextFormatFlags.Left;
            TextRenderer.DrawText(e.Graphics, e.ToolTipText, this.Font, rText, textColor, textFormatFlags);
        }

        /// <summary>
        /// 使い方
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void manualToolStripMenuItem_Click(object sender, EventArgs e)
        {
           app.CommandManager.CancelCurrentCommand();

            // 説明図面ファイルを開く
            string addInFullPath = app.AddInFullPath;// アドインのファイルパス
            string folderPath = Path.GetDirectoryName(addInFullPath) + "\\";
            string filePathName = folderPath + helpFileName;
 
            // 開く
            DocumentCollection documentCollection = app.Documents;
            Document docA = documentCollection.Open(filePathName, true);
            if(docA == null)
                return;

            // 日本語、英語で表示レイヤを変える（非表示のレイヤに説明文を書いておく）
            ApplicationSettings appSetting = app.Application.Settings;
            string languageCode = appSetting.LanguageCode;
            string layerName = "English";
            if(languageCode == "ja")
                layerName = "Japanese";

            Layer rootLayer = docA.LayerTable.RootLayer;
            LayerCollection layers = rootLayer.ChildLayers;
            Layer layer = AddInTools.GetFirstFindedLayerByName(layers, layerName);
            if(layer != null)
            {
                docA.UndoManager.BeginUndoUnit();

                // 英語か日本語のレイヤをカレントにして表示させる
                docA.LayerTable.CurrentLayer = layer;

                docA.UndoManager.EndUndoUnit();
            }

        }

        /// <summary>
        /// このアドインについて
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void aboutThisStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutBox aboutbox = new AboutBox();
            aboutbox.ShowDialog();
        }

        /// <summary>
        /// レイヤグループのプロパティ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layerGroupPropertyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // このレイヤグループのID
            int changeLayerGroupID = GetContextMenuSelectLayerGroupID();
            if(changeLayerGroupID < 0)
                return;

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            LayerTable layerTable = doc.LayerTable;
            Layer layer = layerTable.GetLayerByID(changeLayerGroupID);
            Debug.Assert(layer.IsGroup);

            // プロパティウィンドウにレイヤのプロパティを表示
            ShowLayerProperty(layer);
        }

        /// <summary>
        /// プロパティウィンドウにレイヤのプロパティを表示
        /// </summary>
        /// <param name="layer"></param>
        public void ShowLayerProperty(Layer layer)
        {
            Debug.Assert(layer != null);

            // プロパティウィンドウを表示させる
            app.PropertiesBar.Activate();
            
            // レイヤウィンドウを表示させる
            app.LayerBar.Activate();

            // レイヤを渡してレイヤのプロパティの内容をプロパティウィンドウに表示させる
            app.LayerBar.SelectNode(layer);

            // このアドインのページがレイヤウィンドウに隠れているかもしれないのでアクティブにするなら以下の処理をする
            //app.ActivateMyAddInDockingBar();
        }

        /// <summary>
        /// 部分図レイアウトのプロパティ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void layoutPropertyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // このレイヤグループと連携する部分図レイアウト
            Layout interlockingDrawingLayout = GetContextMenuSelectLayerGroupInterlockingDrawingLayout();
            if(interlockingDrawingLayout == null)
                return;

            // プロパティウィンドウに部分図レイアウトのプロパティを表示
            ShowLayoutProperty(interlockingDrawingLayout);
        }

        /// <summary>
        /// プロパティウィンドウに部分図レイアウトのプロパティを表示
        /// </summary>
        /// <param name="layout"></param>
        public void ShowLayoutProperty(Layout layout)
        {
            Debug.Assert(layout != null);

            // プロパティウィンドウを表示させる
            app.PropertiesBar.Activate();
            
            // 部分図ウィンドウを表示させる
            app.DrawingBar.Activate();

            // 部分図レイアウトを渡して部分図レイアウトのプロパティの内容をプロパティウィンドウに表示させる
            app.DrawingBar.SelectNode(layout);

            // このアドインのページがレイヤウィンドウに隠れているかもしれないのでアクティブにするなら以下の処理をする
            //app.ActivateMyAddInDockingBar();
        }

        /// <summary>
        /// 用紙タブビューをアクティブにする
        /// </summary>
        /// <param name="app"></param>
        private void ActiveFirstPaperTabView()
        {
            if(app.ActiveWindow == null)
                return;

            TabView activeTabView = app.ActiveWindow.ActiveTabView;
            if(activeTabView == null)
                return;

            // 現在用紙ならなにもしない
            if(activeTabView.TabViewType == TabViewType.PaperView)
                return;

            // 他のビューで用紙のものを探す
            TabViewCollection tabViews = app.ActiveWindow.TabViews;
            foreach(TabView tabView in tabViews)
            {
                // 用紙ビューの場合はこれをアクティブにする
                if(tabView.TabViewType == TabViewType.PaperView)
                {
                    tabView.Activate();
                    return;
                }
            }
        }

        /// <summary>
        /// 全ての編集可能な部分図を表示のみにする
        /// (用紙と非表示のものは除く)
        /// </summary>
        private void SetAllPartDrawingStateEditableToVisible()
        {
            if(app.ActiveWindow == null)
                return;

            Document doc = app.ActiveDocument;
            if(doc == null)
                return;

            // 部分図の状態変更
            DrawingCollection drawings = doc.Drawings;
            foreach(Drawing drawing in drawings)
            {
                if(drawing is Paper)
                    continue;

                // 編集可能な部分図を表示のみにする
                if(drawing.State == ShapeState.Editable)
                    drawing.State = ShapeState.Visible;
            }
        }

        /// <summary>
        /// すべての部分図のタブを閉じる
        /// </summary>
        /// <param name="app"></param>
        private void CloseAllPartDrawingView()
        {
            if(app.ActiveWindow == null)
                return;

            TabView activeTabView = app.ActiveWindow.ActiveTabView;
            if(activeTabView == null)
                return;

            TabViewCollection tabViews = app.ActiveWindow.TabViews;
            int tabViewsCount = tabViews.Count;
            for(int i = tabViewsCount - 1; i >= 0; i--)
            {
                TabView tabView = tabViews[i];
                // 部分図タブの場合はこれを閉じる
                if(tabView.TabViewType == TabViewType.PartView)
                    tabView.Close();
            }
        }

        /// <summary>
        /// 指定した部分図IDの部分図タブを閉じる
        /// </summary>
        /// <param name="drawingID"></param>
        private void ClosePartDrawingView(int drawingID)
        {
          if(app.ActiveWindow == null)
                return;

            TabView activeTabView = app.ActiveWindow.ActiveTabView;
            if(activeTabView == null)
                return;

            TabViewCollection tabViews = app.ActiveWindow.TabViews;
            foreach(TabView tabView in tabViews)
            {
                // 部分図タブの場合
                if(tabView.TabViewType == TabViewType.PartView)
                {
                    // 指定した部分図IDの部分図のタブだった場合
                    DrawingView drawingView = (DrawingView)tabView;
                    if(drawingID == drawingView.Drawing.ID)
                    {
                        // 閉じる
                        tabView.Close();
                        return;
                    }
                }
            }
        }

        /// <summary>
        /// 「[表示のみ]のレイヤ、部分図上の図形を選択する」を Off 推奨設定
        /// </summary>
        private void RecommendedCannotSelectShapeInShowOnlyLayerAndPartDrawing()
        {
            // RootPro CAD の設定関係を取得
            ApplicationSettings settings = app.Application.Settings;

            // 「[表示のみ]のレイヤ、部分図上の図形を選択する」がfalseなら何もしない
            if(!settings.SelectVisibleOnlyShape)
                return;

            string msg = Properties.Resources.recommendedMsg_1;
            if(DialogResult.Yes == MessageBox.Show(msg, Properties.Resources.addInProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Information))
            {
                // 「[表示のみ]のレイヤ、部分図上の図形を選択する」を Offに
                settings.SelectVisibleOnlyShape = false;
            }
        }

        /// <summary>
        /// フローティングバーのサイズを合わせる
        /// </summary>
        private void AdjustFloatingBar()
        {
            // If floating, adjust the size of the docking bar
            DockingBarCollection dockingBarCollection = app.DockingBars;
            CustomDockingBar myDockingBar = dockingBarCollection.Find(app.myAddInUniqueName);
            if(myDockingBar != null && myDockingBar.Floating)
            {
                int margin = layerGroup_GroupBox.Bounds.Left;
                int cx = margin * 2 + layerGroup_GroupBox.Bounds.Width;
                int cy = layerGroup_GroupBox.Bounds.Top + layerGroup_GroupBox.Bounds.Height + margin;

                // Adjust FloatingBar
                myDockingBar.AdjustFloatingBar(cx, cy);
            }
        }
    }
}
